#
#  Name:    Mp.py
#
#  Purpose: Python module interface to MX.
#
#  Author:  William Lavender
#
#-----------------------------------------------------------------------
#
#  Copyright 2002-2008 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

import sys, re, types
import MpPrivate

#
#  Initialize the exceptions used by Mp
#

MpPrivate.init_exceptions( sys.modules["Mp"] )

#
#  Initialize the rest of Mp
#

MpPrivate.init()

#
#  Initialize global variable.
#

__program_name__ = None

#=============================================================================
#
#  Mp Module functions
#

def setup_database( filename ):
	#
	#  Create a new record list.
	#

	record_list = RecordList()

	#
	#  Enable plotting
	#

	record_list.set_plot_enable(1)

	#
	#  Read the database in and initialize the corresponding hardware.
	#

	record_list.read_database( filename )

	record_list.finish_database_initialization()

	record_list.initialize_hardware( 0 )

	record_list.unsaved_scans = 0

	return record_list

#
#  Mp output functions
#

def info( message_string ):
	MpPrivate.info( message_string )

def set_info_output_function( function ):
	MpPrivate.set_info_output_function( function )

###

def scanlog_info( message_string ):
	MpPrivate.scanlog_info( message_string )

def set_scanlog_enable( enable_flag ):
	MpPrivate.set_scanlog_enable( enable_flag )

def get_scanlog_enable():
	return MpPrivate.get_scanlog_enable()

###

def warning( message_string ):
	MpPrivate.warning( message_string )

def set_warning_output_function( function ):
	MpPrivate.set_warning_output_function( function )

###

def debug( debug_level, message_string ):
	MpPrivate.debug( debug_level, message_string )

def set_debug_output_function( function ):
	MpPrivate.set_debug_output_function( function )

###

def set_error_output_function( function ):
	MpPrivate.set_error_output_function( function )

###

def info_dialog( text_prompt, gui_prompt, button_label ):
	MpPrivate.info_dialog( text_prompt, gui_prompt, button_label )

def set_info_dialog_function( function ):
	MpPrivate.set_info_dialog_function( function )

###

def info_entry_dialog( text_prompt, gui_prompt, echo_characters ):
	return MpPrivate.info_entry_dialog( text_prompt, \
		gui_prompt, echo_characters )

def set_info_entry_dialog_function( function ):
	MpPrivate.set_info_entry_dialog_function( function )

#
#  Mp user interrupt functions
#

def getch():
	return MpPrivate.getch()

def kbhit():
	return MpPrivate.kbhit()

def user_requested_interrupt():
	return MpPrivate.user_requested_interrupt()

def set_user_interrupt_function( function ):
	MpPrivate.set_user_interrupt_function( function )

def set_scan_pause_request_handler( function ):
	MpPrivate.set_scan_pause_request_handler( function )

#
#  Network functions
#

#
# A comment about Mp.parse_network_field_id():  Obviously,
# Python would be perfectly capable of parsing network field
# ids itself.  The only reason for passing the job to the C
# library is so that we can be completely sure that MX and Mp
# parse the field ids in an identical manner.
#

def parse_network_field_id( network_field_id ):
	return MpPrivate.parse_network_field_id( network_field_id )

#
#  Miscellaneous module functions
#

def get_debug_level():
	return MpPrivate.get_debug_level()

def set_debug_level( debug_level ):
	MpPrivate.set_debug_level( debug_level )

def get_version():
	return MpPrivate.get_version()

def start_py_debugger():
	import pydb
	pydb.set_trace()

def start_c_debugger():
	MpPrivate.start_c_debugger()

def construct_record_object( handle ):
	record_type_string = MpPrivate.get_object_class_name( handle )

	command = "result = %s()" % record_type_string

	exec command

	return result

#=============================================================================
#
#  Mp Classes
#

#
#  Mp image sequence definitions
#

MXT_SQ_ONE_SHOT            = 1
MXT_SQ_CONTINUOUS          = 2
MXT_SQ_MULTIFRAME          = 3
MXT_SQ_CIRCULAR_MULTIFRAME = 4
MXT_SQ_STROBE              = 5
MXT_SQ_BULB                = 6

MXT_SQ_GEOMETRICAL         = 101
MXT_SQ_STREAK_CAMERA       = 102
MXT_SQ_SUBIMAGE            = 103

#
# 2-dimensional image class.
#

class ImageFrame:
	def __init__( self, \
			framesize, image_format, pixel_order, \
			bytes_per_pixel, header_length, image_length ):

		self.handle = MpPrivate.image_init( \
			framesize, image_format, pixel_order, \
			bytes_per_pixel, header_length, image_length )

		self.framesize       = framesize
		self.image_format    = image_format
		self.pixel_order     = pixel_order
		self.bytes_per_pixel = bytes_per_pixel
		self.header_length   = header_length
		self.image_length    = image_length;

	def __del__( self ):
		MpPrivate.image_del( self.handle )

	def get_format_type_from_name( self, name ):
		return MpPrivate.image_get_format_type_from_name( name )

	def get_format_name_from_type( self, type ):
		return MpPrivate.image_get_format_name_from_type( type )

	def get_exposure_time( self ):
		return MpPrivate.image_get_exposure_time( self.handle )

	def get_header_data( self ):
		raise NotYetImplementedError, \
			"FIXME: get_header_data() is not yet implemented."

	def set_header_data( self, header_data ):
		raise NotYetImplementedError, \
			"FIXME: set_header_data() is not yet implemented."

	def get_image_data( self ):
		binary_string = MpPrivate.image_get_image_data( self.handle )

		return binary_string

	def set_image_data( self, image_data ):
		raise NotYetImplementedError, \
			"FIXME: set_image_data() is not yet implemented."

	def copy_frame( self ):
		new_frame = ImageFrame( self.framesize, \
			self.image_format, self.pixel_order, \
			self.bytes_per_pixel, self.header_length, \
			self.image_length )

		header_data = self.get_header_data()

		new_frame.set_header_data( header_data )

		image_data  = self.get_image_data()

		new_frame.set_image_data( image_data )

		return new_frame

	def read_file( self, datafile_type, datafile_name ):
		result = MpPrivate.image_read_file( self.handle, \
					datafile_type, datafile_name )

		(self.framesize, self.image_format, \
			self.pixel_order, self.bytes_per_pixel, \
			self.header_length, self.image_length ) = result

	def write_file( self, datafile_type, datafile_name ):
		MpPrivate.image_write_file( self.handle, \
					datafile_type, datafile_name )

#=============================================================================

#  Callback MX class

MXCBC_NETWORK		= 1
MXCBC_FIELD		= 2

#  Callback MX type

MXCBT_NONE		= 0
MXCBT_VALUE_CHANGED	= 1
MXCBT_POLL		= 2
MXCBT_MOTOR_BACKLASH	= 3
MXCBT_FUNCTION		= 4

#  Callback Python class

class Callback:
	def __init__( self, callback_id, \
			callback_type, callback_function, callback_arguments ):

		self.callback_id = callback_id
		self.callback_type = callback_type
		self.callback_function = callback_function
		self.callback_arguments = callback_arguments

	def __del__( self ):
		try:
			MpPrivate.net_del_callback( self )
		except:
			pass

#=============================================================================

#
#  RecordList class
#

class RecordList:
	def __init__( self ):

		self.unsaved_scans = 0

		# Create a C record list and get the long integer handle that
		# corresponds to the C list head record.

		list_head_handle = MpPrivate.create_record_list()

		list_head_record = construct_record_object( list_head_handle )

		list_head_record.handle = list_head_handle 
		list_head_record.name = \
				MpPrivate.get_record_name( list_head_handle )

		# Store copies of the Python list head object in the Python
		# Mp.RecordList object.

		self.list_head_handle = list_head_handle

		self.list_head_record = list_head_record

	def read_database( self, filename ):

		MpPrivate.read_database( self.list_head_handle, filename )

		# Create a shadow linked list of Mp.Record objects
		# that parallels the MX_RECORD linked list in the
		# C data structures.  This is inserted after the
		# last Mp.Record that already exists in the Python
		# database.

		list_head_record = self.list_head_record
		list_head_handle = self.list_head_handle

		current_record = list_head_record.previous_record
		current_handle = current_record.handle

		while ( 1 ):

			# Find the handle corresponding to the next MX_RECORD
			# in the C database.

			next_handle = \
				MpPrivate.get_next_record( current_handle )

			if ( next_handle == list_head_handle ):
				# We have completed the loop and are back
				# at the start of the C record list.  This
				# means we are done and can now exit the
				# while loop.

				break		

			# Create and fill in a Python Mp.Record object to match
			# the C MX_RECORD.

			next_record = construct_record_object( next_handle )

			next_record.handle = next_handle

			next_record.previous_record = current_record
			next_record.next_record     = list_head_record

			next_record.name = \
				MpPrivate.get_record_name( next_handle )

			# Insert the Python Mp.Record in the Python record list.

			list_head_record.previous_record = next_record
			current_record.next_record = next_record

			current_record = next_record
			current_handle = next_handle

	def finish_database_initialization( self ):
		MpPrivate.finish_database_initialization( \
						self.list_head_handle )

	def initialize_hardware( self, inithw_flags ):
		MpPrivate.initialize_hardware( self.list_head_handle, \
						inithw_flags )

	def shutdown_hardware( self ):
		MpPrivate.shutdown_hardware( self.list_head_handle )

	def create_record_from_description( self, description ):
		# Create a C MX_RECORD and get back a handle for it.

		handle = MpPrivate.create_record_from_description( \
					self.list_head_handle, description )

		# Construct a Python Mp.Record that corresponds to it.

		record = construct_record_object( handle )

		record.handle = handle
		record.name = MpPrivate.get_record_name( handle )

		# The C database generally adds records to the end of the
		# circular MX_RECORD list, so we should prepare to do the same.

		list_head_record = self.list_head_record
		previous_record  = list_head_record.previous_record

		record.next_record     = list_head_record
		record.previous_record = previous_record

		# Now insert the Mp.Record into the circular list.

		list_head_record.previous_record = record
		previous_record.next_record      = record

		# Return the record we have just created.

		return record

	def get_record( self, record_name ):

		list_head_record = self.list_head_record

		if ( list_head_record.name == record_name ):
			return list_head_record

		current_record = list_head_record.next_record

		# Search through the record list looking for the requested
		# record.

		while ( current_record != list_head_record ):

			# If we find it, then return.

			if ( current_record.name == record_name ):
				return current_record

			current_record = current_record.next_record

		# If we get here, then we did not find the record
		# and must return an error.

		error_string = \
			"The record '%s' does not exist." % ( record_name )

		raise Not_Found_Error, error_string

	def get_multiple_records( self, mx_superclass, mx_class, mx_type, \
					reverse_order ):

		if ( reverse_order == 0 ):
			current_record = self.list_head_record.next_record
		else:
			current_record = self.list_head_record.previous_record

		python_record_list = []

		while ( current_record != self.list_head_record ):

			current_name = current_record.name

			current_superclass = \
				current_record.get_field( "mx_superclass" )

			match = re.search( mx_superclass, current_superclass )

			if match:
				current_class = \
					current_record.get_field( "mx_class" )

				match = re.search( mx_class, current_class )

				if match:
					current_type = \
					  current_record.get_field( "mx_type" )

					match = re.search( mx_type, \
								current_type )
					if match:
						python_record_list.append( \
								current_name )

			if ( reverse_order == 0 ):
				current_record = current_record.next_record
			else:
				current_record = current_record.previous_record

		return python_record_list

	def add_handle_to_record_list( self, handle ):
		# FIXME: This should be used by create_record_from_description.

		# Construct a Python Mp.Record that corresponds to the handle.

		record = construct_record_object( handle )

		record.handle = handle
		record.name = MpPrivate.get_record_name( handle )

		# The C database generally adds records to the end of the
		# circular MX_RECORD list, so we should do the same for
		# the matching Python list.

		list_head_record = self.list_head_record
		previous_record  = list_head_record.previous_record

		record.next_record     = list_head_record
		record.previous_record = previous_record

		# Now insert the Mp.Record into the circular list.

		list_head_record.previous_record = record
		previous_record.next_record      = record

		# Return the record that we have just created.

		return record

	def connect_to_mx_server( self, server_name, server_port, server_flags):
		# Connect to the MX server and get a handle for the new
		# server record.

		server_handle = MpPrivate.connect_to_mx_server( \
					self.list_head_handle, \
					server_name, server_port, server_flags )

		# Insert the handle into the record list.

		server_record = self.add_handle_to_record_list( server_handle )

		return server_record

	def wait_for_messages( self, timeout ):
		return MpPrivate.wait_for_messages( self.list_head_handle, \
							timeout )

	def set_network_debug( self, debug ):
		return MpPrivate.set_network_debug( self.list_head_handle, \
							debug )

	def plotting_is_enabled( self ):
		return MpPrivate.plotting_is_enabled( self.list_head_handle )

	def set_plot_enable( self, plot_enable_value ):
		MpPrivate.set_plot_enable( self.list_head_handle, \
						plot_enable_value )

	def set_program_name( self, program_name ):
		MpPrivate.set_program_name( self.list_head_handle, \
						program_name )

	def log_open( self ):
		MpPrivate.log_open( self.list_head_handle )

	def log_close( self ):
		MpPrivate.log_close( self.list_head_handle )

#
#  Record class
#

class Record:
	def __init__( self ):
		self.previous_record = self
		self.next_record = self

	def create_description_from_record( self ):
		return MpPrivate.create_description_from_record( self.handle )

	def finish_record_initialization( self ):
		MpPrivate.finish_record_initialization( self.handle )

	def open_hardware( self ):
		MpPrivate.open_hardware( self.handle )

	def resynchronize( self ):
		MpPrivate.resynchronize( self.handle )

	def delete_record( self ):
		previous_record = self.previous_record
		next_record = self.next_record

		# Remove this record from the Python record list.

		next_record.previous_record = previous_record
		previous_record.next_record = next_record

		# Then, delete the C record.

		MpPrivate.delete_record( self.handle )

		# Finally, delete the Python record.

		del self

	def get_next_record( self ):
		return self.next_record

	def get_previous_record( self ):
		return self.previous_record

	def get_field( self, field_name ):
		return MpPrivate.get_field( self.handle, field_name )

	def set_field( self, field_name, string_value ):
		MpPrivate.set_field( self.handle, field_name, string_value )

#
#  CAMAC interface class
#

class CAMAC(Record):
	def command( self, slot, subaddress, function_code, data, \
								mode="normal" ):
		return MpPrivate.camac_command( self.handle, \
				slot, subaddress, function_code, data, mode )

	def get_lam_status( self ):
		return MpPrivate.camac_get_lam_status( self.handle )

	def controller_command( self, command ):
		MpPrivate.camac_controller_command( self.handle, command )

#
#  GPIB interface class
#

class GPIB(Record):
	def getline( self, address, maximum_length=1024 ):
		return MpPrivate.gpib_getline( self.handle, \
						address, maximum_length )

	def putline( self, address, line ):
		MpPrivate.gpib_putline( self.handle, address, line )

	def interface_clear( self ):
		MpPrivate.gpib_interface_clear( self.handle )

	def device_clear( self ):
		MpPrivate.gpib_device_clear( self.handle )

	def selective_device_clear( self, address ):
		MpPrivate.gpib_selective_device_clear( self.handle, address )

	def local_lockout( self ):
		MpPrivate.gpib_local_lockout( self.handle )

	def remote_enable( self, address ):
		MpPrivate.gpib_remote_enable( self.handle, address )

	def go_to_local( self, address ):
		MpPrivate.gpib_go_to_local( self.handle, address )

	def trigger( self, address ):
		MpPrivate.gpib_trigger( self.handle, address )

	def wait_for_service_request( self, timeout ):
		MpPrivate.gpib_wait_for_service_request( self.handle, timeout )

	def serial_poll( self, address ):
		return MpPrivate.gpib_serial_poll( self.handle, address )

	def serial_poll_disable( self ):
		MpPrivate.gpib_serial_poll_disable( self.handle )

#
#  MODBUS interface class
#

class MODBUS(Record):
	def read_coils( self, starting_address, num_coils ):
		return MpPrivate.modbus_read_coils( self.handle, \
					starting_address, num_coils )

	def read_discrete_inputs( self, starting_address, num_inputs ):
		return MpPrivate.modbus_read_discrete_inputs( self.handle, \
					starting_address, num_inputs )

	def read_holding_registers( self, starting_address, num_registers ):
		return MpPrivate.modbus_read_holding_registers( self.handle, \
					starting_address, num_registers )

	def read_input_registers( self, starting_address, num_registers ):
		return MpPrivate.modbus_read_input_registers( self.handle, \
					starting_address, num_registers )

	def write_single_coil( self, coil_address, coil_value ):
		MpPrivate.modbus_write_single_coil( self.handle, \
					coil_address, coil_value )

	def write_single_register( self, register_address, register_value ):
		MpPrivate.modbus_write_single_register( self.handle, \
					register_address, register_value )

	def read_exception_status( self ):
		return MpPrivate.modbus_read_exception_status( self.handle )

	def get_comm_event_counter( self ):
		return MpPrivate.modbus_get_comm_event_counter( self.handle )

	def write_multiple_coils( self, starting_address, coil_value_tuple ):
		MpPrivate.modbus_write_multiple_coils( self.handle, \
					starting_address, coil_value_tuple )

	def write_multiple_registers( self, \
				starting_address, register_value_tuple ):
		MpPrivate.modbus_write_multiple_registers( self.handle, \
				starting_address, register_value_tuple )

	def mask_write_register( self, register_address, and_mask, or_mask ):
		MpPrivate.modbus_mask_write_register( self.handle, \
				register_address, and_mask, or_mask )

	def read_write_multiple_registers( self, \
			read_starting_address, num_registers_to_read, \
			write_starting_address, write_register_tuple ):
		return MpPrivate.modbus_read_write_multiple_registers( \
			self.handle, \
			read_starting_address, num_registers_to_read, \
			write_starting_address, write_register_tuple )

	def read_fifo_queue( self, fifo_pointer_address, max_fifo_values ):
		return MpPrivate.modbus_read_fifo_queue( self.handle, \
			fifo_pointer_address, max_fifo_values )
#
#  Port I/O interface class
#

class Portio(Record):
	def inp8( self, port_number ):
		return MpPrivate.portio_inp8( self.handle, port_number )

	def inp16( self, port_number ):
		return MpPrivate.portio_inp16( self.handle, port_number )

	def outp8( self, port_number, value ):
		MpPrivate.portio_outp8( self.handle, port_number, value )

	def outp16( self, port_number, value ):
		MpPrivate.portio_outp16( self.handle, port_number, value )

#
#  RS-232 interface class
#

class RS232(Record):
	def getchar( self ):
		return MpPrivate.rs232_getchar( self.handle )

	def putchar( self, c ):
		MpPrivate.rs232_putchar( self.handle, c )

	def read( self, num_bytes ):
		return MpPrivate.rs232_read( self.handle, num_bytes )

	def write( self, buffer ):
		MpPrivate.rs232_write( self.handle, buffer )

	def getline( self, maximum_length=1024 ):
		return MpPrivate.rs232_getline( self.handle, maximum_length )

	def putline( self, line ):
		MpPrivate.rs232_putline( self.handle, line )

	def num_input_bytes_available( self ):
		return MpPrivate.rs232_num_input_bytes_available( self.handle )

	def discard_unread_input( self ):
		MpPrivate.rs232_discard_unread_input( self.handle )

	def discard_unwritten_output( self ):
		MpPrivate.rs232_discard_unwritten_output( self.handle )

	def get_signal_state( self ):
		return MpPrivate.rs232_get_signal_state( self.handle )

	def set_signal_state( self, signal_state ):
		MpPrivate.rs232_set_signal_state( self.handle, signal_state )

#
#  USB interface class
#

class USB(Record):
	def enumerate_devices( self ):
		MpPrivate.usb_enumerate_devices( self.handle )

	def find_device_by_order( self, vendor_id, product_id, order_number, \
			configuration_number, interface_number, \
			alternate_interface_number, quiet_flag ):

		device_handle = MpPrivate.usb_find_device_by_order( \
			self.handle, vendor_id, product_id, order_number, \
			configuration_number, interface_number, \
			alternate_interface_number, quiet_flag )

		return USBDevice( self, device_handle )

	def find_device_by_serial_number( self, \
			vendor_id, product_id, serial_number, \
			configuration_number, interface_number, \
			alternate_interface_number, quiet_flag ):

		device_handle = MpPrivate.usb_find_device_by_order( \
			self.handle, vendor_id, product_id, serial_number, \
			configuration_number, interface_number, \
			alternate_interface_number, quiet_flag )

		return USBDevice( self, device_handle )

#
#  USBDevice class
#

class USBDevice:
	def __init__( self, bus, handle ):
		self.bus = bus
		self.handle = handle

	def delete_device( self ):
		MpPrivate.usb_delete_device( self.bus.handle, self.handle )

	def reset_device( self ):
		MpPrivate.usb_reset_device( self.bus.handle, self.handle )

	def control_transfer( self, \
			request_type, request, value, \
			index_or_offset, write_buffer, \
			num_bytes_to_read, timeout ):

		return MpPrivate.usb_control_transfer( \
			self.bus.handle, self.handle, \
			request_type, request, value, \
			index_or_offset, write_buffer, \
			num_bytes_to_read, timeout )

	def bulk_read( self, endpoint_number, num_bytes_to_read, timeout ):

		return MpPrivate.usb_bulk_read( self.bus.handle, self.handle, \
			endpoint_number, num_bytes_to_read, timeout )

	def bulk_write( self, endpoint_number, write_buffer, timeout ):

		MpPrivate.usb_bulk_write( self.bus.handle, self.handle, \
			endpoint_number, write_buffer, timeout )

#
#  VME interface class
#

class VME(Record):
	def in8( self, crate, address_mode, address ):
		return MpPrivate.vme_in8( self.handle, \
				crate, address_mode, address )

	def in16( self, crate, address_mode, address ):
		return MpPrivate.vme_in16( self.handle, \
				crate, address_mode, address )

	def in32( self, crate, address_mode, address ):
		return MpPrivate.vme_in32( self.handle, \
				crate, address_mode, address )

	def out8( self, crate, address_mode, address, value ):
		MpPrivate.vme_out8( self.handle, \
				crate, address_mode, address, value )

	def out16( self, crate, address_mode, address, value ):
		MpPrivate.vme_out16( self.handle, \
				crate, address_mode, address, value )

	def out32( self, crate, address_mode, address, value ):
		MpPrivate.vme_out32( self.handle, \
				crate, address_mode, address, value )

	def multi_in8( self, crate, address_mode, address_increment, \
			address, num_values ):
		return MpPrivate.vme_multi_in8( self.handle, crate, \
			address_mode, address_increment, address, num_values )

	def multi_in16( self, crate, address_mode, address_increment, \
			address, num_values ):
		return MpPrivate.vme_multi_in16( self.handle, crate, \
			address_mode, address_increment, address, num_values )

	def multi_in32( self, crate, address_mode, address_increment, \
			address, num_values ):
		return MpPrivate.vme_multi_in32( self.handle, crate, \
			address_mode, address_increment, address, num_values )

	def multi_out8( self, crate, address_mode, address_increment, \
			address, values_tuple ):
		MpPrivate.vme_multi_out8( self.handle, crate, \
			address_mode, address_increment, address, values_tuple )

	def multi_out16( self, crate, address_mode, address_increment, \
			address, values_tuple ):
		MpPrivate.vme_multi_out16( self.handle, crate, \
			address_mode, address_increment, address, values_tuple )

	def multi_out32( self, crate, address_mode, address_increment, \
			address, values_tuple ):
		MpPrivate.vme_multi_out32( self.handle, crate, 
			address_mode, address_increment, address, values_tuple )

	def get_read_address_increment( self, crate ):
		return MpPrivate.vme_get_read_address_increment( self.handle, \
									crate )

	def set_read_address_increment( self, crate, address_increment ):
		MpPrivate.vme_set_read_address_increment( self.handle, crate, \
							address_increment )

	def get_write_address_increment( self, crate ):
		return MpPrivate.vme_get_write_address_increment( self.handle, \
									crate )

	def set_write_address_increment( self, crate, address_increment ):
		MpPrivate.vme_set_write_address_increment( self.handle, crate, \
							address_increment )

#
#  Amplifier class
#

class Amplifier(Record):

	def get_gain( self ):
		return MpPrivate.amplifier_get_gain( self.handle )

	def set_gain( self, value ):
		MpPrivate.amplifier_set_gain( self.handle, value )

	def get_offset( self ):
		return MpPrivate.amplifier_get_offset( self.handle )

	def set_offset( self, value ):
		MpPrivate.amplifier_set_offset( self.handle, value )

	def get_time_constant( self ):
		return MpPrivate.amplifier_get_time_constant( self.handle )

	def set_time_constant( self, value ):
		MpPrivate.amplifier_set_time_constant( self.handle, value )

#
#  Analog input class
#

class AnalogInput(Record):

	def read( self ):
		return MpPrivate.analog_input_read( self.handle )

	def read_raw_long( self ):
		return MpPrivate.analog_input_read_raw_long( self.handle )

	def read_raw_double( self ):
		return MpPrivate.analog_input_read_raw_double( self.handle )

	def get_dark_current( self ):
		return MpPrivate.analog_input_get_dark_current( self.handle )

	def set_dark_current( self, dark_current ):
		MpPrivate.analog_input_set_dark_current( self.handle, \
							dark_current )

#
#  Analog output class
#

class AnalogOutput(Record):

	def read( self ):
		return MpPrivate.analog_output_read( self.handle )

	def read_raw_long( self ):
		return MpPrivate.analog_output_read_raw_long( self.handle )

	def read_raw_double( self ):
		return MpPrivate.analog_output_read_raw_double( self.handle )

	def write( self, value ):
		MpPrivate.analog_output_write( self.handle, value )

	def write_raw_long( self, value ):
		MpPrivate.analog_output_write_raw_long( self.handle, value )

	def write_raw_double( self, value ):
		MpPrivate.analog_output_write_raw_double( self.handle, value )

#
#  Area detector definitions
#

MXFT_AD_IMAGE_FRAME        = 0x0
MXFT_AD_MASK_FRAME         = 0x1
MXFT_AD_BIAS_FRAME         = 0x2
MXFT_AD_DARK_CURRENT_FRAME = 0x4
MXFT_AD_FLOOD_FIELD_FRAME  = 0x8

#
#  Area detector class
#

class AreaDetector(Record):
	# FIXME: Put property stuff here.

	#-----------------------------------------------------------------

	def get_image_format( self ):
		return MpPrivate.area_detector_get_image_format( self.handle )

	def set_image_format( self, format ):
		MpPrivate.area_detector_set_image_format( self.handle, format )

	def get_maximum_framesize( self ):
		return MpPrivate.area_detector_get_maximum_framesize( \
								self.handle )

	def get_framesize( self ):
		return MpPrivate.area_detector_get_framesize( self.handle )

	def set_framesize( self, x_framesize, y_framesize ):
		MpPrivate.area_detector_get_framesize( self.handle, \
							framesize_tuple )

	def get_binsize( self ):
		return MpPrivate.area_detector_get_binsize( self.handle )

	def set_binsize( self, x_binsize, y_binsize ):
		MpPrivate.area_detector_get_binsize( self.handle, \
							binsize_tuple )

	def get_roi( self, roi_number ):
		return MpPrivate.area_detector_get_roi( self.handle, \
							roi_number )

	def set_roi( self, roi_number, roi_tuple ):
		MpPrivate.area_detector_set_roi( self.handle, \
						roi_number, roi_tuple )

	# FIXME: Insert subframe or hw roi stuff here.

	def get_bytes_per_frame( self ):
		return MpPrivate.area_detector_get_bytes_per_frame(self.handle)

	def get_bytes_per_pixel( self ):
		return MpPrivate.area_detector_get_bytes_per_pixel(self.handle)

	def get_bits_per_pixel( self ):
		return MpPrivate.area_detector_get_bits_per_pixel(self.handle)

	def get_correction_flags( self ):
		return MpPrivate.area_detector_get_correction_flags(self.handle)

	def set_correction_flags( self, format ):
		MpPrivate.area_detector_set_correction_flags( self.handle, \
								format )

	def measure_correction_frame( self, correction_measurement_type, \
					correction_measurement_time, \
					num_correction_measurements ):

		MpPrivate.area_detector_measure_correction_frame( self.handle, \
						correction_measurement_type, \
						correction_measurement_time, \
						num_correction_measurements )

	def measure_dark_current_frame( self, correction_measurement_time, \
					num_correction_measurements ):

		MpPrivate.area_detector_measure_correction_frame( self.handle, \
						MXFT_AD_DARK_CURRENT_FRAME, \
						correction_measurement_time, \
						num_correction_measurements )

	def measure_flood_field_frame( self, correction_measurement_time, \
					num_correction_measurements ):

		MpPrivate.area_detector_measure_correction_frame( self.handle, \
						MXFT_AD_FLOOD_FIELD_FRAME, \
						correction_measurement_time, \
						num_correction_measurements )

	def get_use_scaled_dark_current_flag( self ):
		return \
		    MpPrivate.area_detector_get_use_scaled_dark_current_flag( \
								self.handle )

	def set_use_scaled_dark_current_flag( self, format ):
		MpPrivate.area_detector_set_use_scaled_dark_current_flag( \
					self.handle, use_scaled_dark_current )

	#-----------------------------------------------------------------

	def get_detector_readout_time( self ):
		return \
		  MpPrivate.area_detector_get_detector_readout_time(self.handle)

	def get_total_sequence_time( self ):
		return \
		  MpPrivate.area_detector_get_total_sequence_time( self.handle )

	def get_sequence_parameters( self ):
		return \
	    MpPrivate.area_detector_get_sequence_parameters( self.handle )

	def set_sequence_parameters( self, seq_type, seq_params ):
		MpPrivate.area_detector_set_sequence_parameters( \
			self.handle, seq_type, seq_params )

	def set_one_shot_mode( self, exposure_time ):
		seq_params = ( exposure_time, )

		self.set_sequence_parameters( MXT_SQ_ONE_SHOT, seq_params )

	def set_continuous_mode( self, exposure_time ):
		seq_params = ( exposure_time, )

		self.set_sequence_parameters( MXT_SQ_CONTINUOUS, seq_params )

	def set_multiframe_mode( self, num_frames, exposure_time, frame_time ):
		seq_params = ( num_frames, exposure_time, frame_time )

		self.set_sequence_parameters( MXT_SQ_MULTIFRAME, seq_params )

	def set_circular_multiframe_mode( self, \
				num_frames, exposure_time, frame_time ):

		seq_params = ( num_frames, exposure_time, frame_time )

		self.set_sequence_parameters( \
					MXT_SQ_CIRCULAR_MULTIFRAME, seq_params )

	def set_strobe_mode( self, num_frames, exposure_time ):
		seq_params = ( num_frames, exposure_time )

		self.set_sequence_parameters( MXT_SQ_STROBE, seq_params )

	def set_bulb_mode( self, num_frames ):
		seq_params = ( num_frames, )

		self.set_sequence_parameters( MXT_SQ_BULB, seq_params )

	def set_geometrical_mode( self, num_frames, exposure_time, gap_time, \
					exposure_multiplier, gap_multiplier ):

		seq_params = ( num_frames, exposure_time, gap_time, \
				exposure_multiplier, gap_multiplier )

		self.set_sequence_parameters( MXT_SQ_GEOMETRICAL, seq_params )

	#-----------------------------------------------------------------

	def get_trigger_mode( self ):
		return MpPrivate.area_detector_get_trigger_mode( self.handle )

	def set_trigger_mode( self, trigger_mode ):
		MpPrivate.area_detector_set_trigger_mode( self.handle, \
								trigger_mode )

	def arm( self ):
		MpPrivate.area_detector_arm( self.handle )

	def trigger( self ):
		MpPrivate.area_detector_trigger( self.handle )

	def start( self ):
		MpPrivate.area_detector_start( self.handle )

	def stop( self ):
		MpPrivate.area_detector_stop( self.handle )

	def abort( self ):
		MpPrivate.area_detector_abort( self.handle )

	def is_busy( self ):
		return MpPrivate.area_detector_is_busy( self.handle )

	def get_last_frame_number( self ):
		return \
		    MpPrivate.area_detector_get_last_frame_number( self.handle )

	def get_total_num_frames( self ):
		return \
		    MpPrivate.area_detector_get_total_num_frames( self.handle )

	def get_status( self ):
		return MpPrivate.area_detector_get_status( self.handle )

	def get_extended_status( self ):
		return MpPrivate.area_detector_get_extended_status(self.handle)

	def setup_frame( self, image_frame ):
		if ( image_frame == None ):
			raise NotYetImplementedError, \
				"FIXME: Using setup_frame() with a None " \
				"image_frame is not yet implemented."

		MpPrivate.area_detector_setup_frame( \
					self.handle, image_frame.handle )
		return image_frame

	def readout_frame( self, frame_number ):
		MpPrivate.area_detector_readout_frame(self.handle, frame_number)

	def correct_frame( self ):
		MpPrivate.area_detector_correct_frame( self.handle )

	def transfer_frame( self, frame_type, image_frame ):
		if ( image_frame == None ):
			image_frame_handle = -1
		else:
			image_frame_handle = image_frame.handle

		MpPrivate.area_detector_transfer_frame( self.handle, \
						frame_type, image_frame_handle )

	def load_frame( self, frame_type, frame_filename ):
		MpPrivate.area_detector_load_frame( self.handle, \
						frame_type, frame_filename )

	def save_frame( self, frame_type, frame_filename ):
		MpPrivate.area_detector_save_frame( self.handle, \
						frame_type, frame_filename )

	def copy_frame( self, source_frame_type, destination_frame_type ):
		MpPrivate.area_detector_copy_frame( self.handle, \
				source_frame_type, destination_frame_type )

	def get_frame( self, frame_number, image_frame ):
		if ( image_frame == None ):
			raise NotYetImplementedError, \
				"FIXME: Using get_frame() with a None " \
				"image_frame is not yet implemented."

		MpPrivate.area_detector_get_frame( self.handle, \
					frame_number, image_frame.handle )

		return image_frame

	# FIXME: Insert get_roi_frame here.

#
#  Autoscale class
#

class Autoscale(Record):
	def read_monitor( self ):
		return MpPrivate.autoscale_read_monitor( self.handle )

	def get_change_request( self ):
		return MpPrivate.autoscale_get_change_request( self.handle )

	def change_control( self, change_control ):
		MpPrivate.autoscale_change_control( self.handle, \
							change_control )

	def get_offset_index( self ):
		return MpPrivate.autoscale_get_offset_index( self.handle )

	def set_offset_index( self, monitor_offset_index ):
		MpPrivate.autoscale_set_offset_index( self.handle, \
							monitor_offset_index )

	def get_offsets( self ):
		return MpPrivate.autoscale_get_offsets( self.handle )
					
	def set_offsets( self, monitor_offsets ):
		MpPrivate.autoscale_set_offsets( self.handle, monitor_offsets )

	def get_limits( self ):
		return MpPrivate.autoscale_get_limits( self.handle )

	def set_limits( self, limits_tuple ):
		"""
		limits_tuple includes 
		( low_limit, high_limit, low_deadband, high_deadband )
		"""
		MpPrivate.autoscale_set_limits( self.handle, limits_tuple )

	def compute_dynamic_limits( self ):
		MpPrivate.autoscale_compute_dynamic_limits( self.handle )

#
#  Digital input class
#

class DigitalInput(Record):

	def read( self ):
		return MpPrivate.digital_input_read( self.handle )

#
#  Digital output class
#

class DigitalOutput(Record):

	def read( self ):
		return MpPrivate.digital_output_read( self.handle )

	def write( self, value ):
		MpPrivate.digital_output_write( self.handle, value )

	def pulse( self, polarity, busy_wait, pulse_duration_in_microseconds ):
		MpPrivate.digital_output_pulse( self.handle, \
			polarity, busy_wait, pulse_duration_in_microseconds )
#
#  Encoder class
#

class Encoder(Record):
	def get_overflow_status( self ):
		return MpPrivate.encoder_get_overflow_status( self.handle )

	def reset_overflow_status( self ):
		MpPrivate.encoder_reset_overflow_status( self.handle )

	def read( self ):
		return MpPrivate.encoder_read( self.handle )

	def write( self, value ):
		return MpPrivate.encoder_write( self.handle, value )

#
#  Multichannel analyzer class
#

class MCA(Record):
	def start( self ):
		MpPrivate.mca_start( self.handle )

	def stop( self ):
		MpPrivate.mca_stop( self.handle )

	def read( self ):
		return MpPrivate.mca_read( self.handle )

	def clear( self ):
		MpPrivate.mca_clear( self.handle )

	def is_busy( self ):
		return MpPrivate.mca_is_busy( self.handle )

	def start_without_preset( self ):
		MpPrivate.mca_start_without_preset( self.handle )

	def start_with_preset( self, preset_type, preset_value ):
		MpPrivate.mca_start_without_preset( self.handle, \
						preset_type, preset_value )

	def start_for_preset_live_time( self, preset_live_time ):
		MpPrivate.mca_start_for_preset_live_time( self.handle, \
							preset_live_time )

	def start_for_preset_real_time( self, preset_real_time ):
		MpPrivate.mca_start_for_preset_real_time( self.handle, \
							preset_real_time )

	def start_for_preset_count( self, preset_count ):
		MpPrivate.mca_start_for_preset_count( self.handle, \
							preset_count )

	def get_preset_type( self ):
		return MpPrivate.mca_get_preset_type( self.handle )

	def set_preset_type( self, preset_type ):
		MpPrivate.mca_set_preset_type( self.handle, preset_type )

	def get_preset_count_region( self ):
		return MpPrivate.mca_get_preset_count_region( self.handle )

	def set_preset_count_region( self, preset_count_region ):
		MpPrivate.mca_set_preset_count_region( self.handle, \
							preset_count_region )

	def get_roi( self, roi_number ):
		return MpPrivate.mca_get_roi( self.handle, roi_number )

	def set_roi( self, roi_number, roi_tuple ):
		MpPrivate.mca_set_roi( self.handle, roi_number, roi_tuple )

	def get_roi_integral( self, roi_number ):
		return MpPrivate.mca_get_roi_integral( self.handle, roi_number )

	def get_rois( self ):
		return MpPrivate.mca_get_rois( self.handle )

	def set_rois( self, rois_tuple ):
		MpPrivate.mca_set_rois( self.handle, rois_tuple )

	def get_roi_integrals( self ):
		return MpPrivate.mca_get_roi_integrals( self.handle, \
							 roi_integrals_tuple )

	def get_num_channels( self ):
		return MpPrivate.mca_get_num_channels( self.handle )

	def set_num_channels( self, num_channels ):
		MpPrivate.mca_set_num_channels( self.handle, num_channels )

	def get_channel( self ):
		return MpPrivate.mca_get_channel( self.handle, channel_number )

	def get_real_time( self ):
		return MpPrivate.mca_get_real_time( self.handle )

	def get_live_time( self ):
		return MpPrivate.mca_get_live_time( self.handle )

	def get_counts( self ):
		return MpPrivate.mca_get_counts( self.handle )

	def get_soft_roi( self, soft_roi_number ):
		return MpPrivate.mca_get_soft_roi( self.handle, \
							soft_roi_number )

	def set_soft_roi( self, soft_roi_number, soft_roi_tuple ):
		MpPrivate.mca_set_soft_roi( self.handle, \
					 soft_roi_number, soft_roi_tuple )

	def get_soft_roi_integral( self, soft_roi_number ):
		return MpPrivate.mca_get_soft_roi_integral( self.handle, \
							 soft_roi_number )

	def get_soft_roi_integrals( self ):
		return MpPrivate.mca_get_soft_roi_integrals( self.handle, \
							 roi_integrals_tuple )

	def get_energy_scale( self ):
		return MpPrivate.mca_get_energy_scale( self.handle )

	def set_energy_scale( self, energy_scale ):
		MpPrivate.mca_set_energy_scale( self.handle, energy_scale )

	def get_energy_offset( self ):
		return MpPrivate.mca_get_energy_offset( self.handle )

	def set_energy_offset( self, energy_offset ):
		MpPrivate.mca_set_energy_offset( self.handle, energy_offset )

	def get_energy_axis( self ):
		return MpPrivate.mca_get_energy_axis( self.handle )

#
#  Multichannel encoder class
#

class MCE(Record):
	def get_overflow_status( self ):
		return MpPrivate.mce_get_overflow_status( self.handle )

	def reset_overflow_status( self ):
		MpPrivate.mce_reset_overflow_status( self.handle )

	def read( self ):
		return MpPrivate.mce_read( self.handle )

	def get_current_num_values( self ):
		return MpPrivate.mce_get_current_num_values( self.handle )

#
#  Multichannel scaler class
#

class MCS(Record):
	def start( self ):
		MpPrivate.mcs_start( self.handle )

	def stop( self ):
		MpPrivate.mcs_stop( self.handle )

	def clear( self ):
		MpPrivate.mcs_clear( self.handle )

	def is_busy( self ):
		""" Returns busy flag """
		return MpPrivate.mcs_is_busy( self.handle )

	def read_all( self ):
		"""
		Returns a 2-D mcs_data tuple.
		"""
		return MpPrivate.mcs_read_all( self.handle )

	def read_scaler( self, scaler_index ):
		"""
		Returns a 1-D scaler_data tuple.
		"""
		return MpPrivate.mcs_read_scaler( self.handle, scaler_index )

	def read_measurement( self, measurement_index ):
		"""
		Returns a 1-D measurement_data tuple.
		"""
		return MpPrivate.mcs_read_measurement( self.handle, \
							measurement_index )

	def read_timer( self ):
		"""
		Returns a 1-D timer_data tuple.
		"""
		return MpPrivate.mcs_read_timer( self.handle )

	def get_mode( self ):
		""" Returns the counter mode setting """
		return MpPrivate.mcs_get_mode( self.handle )

	def set_mode( self, mode ):
		MpPrivate.mcs_set_mode( self.handle, mode )

	def get_measurement_time( self ):
		""" Returns measurement time per point """
		return MpPrivate.mcs_get_measurement_time( self.handle )

	def set_measurement_time( self, measurement_time ):
		MpPrivate.mcs_set_measurement_time( self.handle, \
							measurement_time )

	def get_measurement_counts( self ):
		""" Returns measurement counts per point """
		return MpPrivate.mcs_get_measurement_counts( self.handle )

	def set_measurement_counts( self, measurement_counts ):
		MpPrivate.mcs_set_measurement_counts( self.handle, \
							measurement_counts )

	def get_num_measurements( self ):
		""" Returns the number of measurements """
		return MpPrivate.mcs_get_num_measurements( self.handle )

	def set_num_measurements( self, num_measurements ):
		MpPrivate.mcs_set_num_measurements( self.handle, \
							num_measurements )

#
#  Motor class
#

class Motor(Record):

	tweak_distance = 0.0

	def is_busy( self ):
		return MpPrivate.motor_is_busy( self.handle )

	def move_absolute( self, new_destination ):
		MpPrivate.motor_move_absolute( self.handle, new_destination )

	def move_relative( self, relative_position ):
		MpPrivate.motor_move_relative( self.handle, relative_position )

	def move_absolute_with_flags( self, new_destination, flags ):
		MpPrivate.motor_move_absolute_with_flags( self.handle, \
						new_destination, flags )

	def move_relative_with_flags( self, relative_position, flags ):
		MpPrivate.motor_move_relative_with_flags( self.handle, \
						relative_position, flags )

	def wait_for_motor_stop( self, flags ):
		MpPrivate.motor_wait_for_motor_stop( self.handle, flags )

	def get_position( self ):
		return MpPrivate.motor_get_position( self.handle )

	def set_position( self, new_set_position ):
		MpPrivate.motor_set_position( self.handle, new_set_position )

	def soft_abort( self ):
		MpPrivate.motor_soft_abort( self.handle )

	def immediate_abort( self ):
		MpPrivate.motor_immediate_abort( self.handle )

	def positive_limit_hit( self ):
		return MpPrivate.motor_positive_limit_hit( self.handle )

	def negative_limit_hit( self ):
		return MpPrivate.motor_negative_limit_hit( self.handle )

	def find_home_position( self, direction ):
		MpPrivate.motor_find_home_position( self.handle, direction )

	def constant_velocity_move( self, direction ):
		MpPrivate.motor_constant_velocity_move( self.handle, direction )

	def get_status( self ):
		return MpPrivate.motor_get_status( self.handle )

	def get_extended_status( self ):
		return MpPrivate.motor_get_extended_status( self.handle )

	def get_speed( self ):
		return MpPrivate.motor_get_speed( self.handle )

	def set_speed( self, speed ):
		MpPrivate.motor_set_speed( self.handle, speed )

	def get_base_speed( self ):
		return MpPrivate.motor_get_base_speed( self.handle )

	def set_base_speed( self, base_speed ):
		MpPrivate.motor_set_base_speed( self.handle, base_speed )

	def get_maximum_speed( self ):
		return MpPrivate.motor_get_maximum_speed( self.handle )

	def set_maximum_speed( self, maximum_speed ):
		MpPrivate.motor_set_maximum_speed( self.handle, maximum_speed )

	def save_speed( self ):
		MpPrivate.motor_save_speed( self.handle )

	def restore_speed( self ):
		MpPrivate.motor_restore_speed( self.handle )

	def set_speed_between_positions( self, \
			position1, position2, time_for_move ):
		MpPrivate.motor_set_speed_between_positions( self.handle, \
			position1, position2, time_for_move )

	def get_acceleration_type( self ):
		return MpPrivate.motor_get_acceleration_type( self.handle )

	def get_raw_acceleration_parameters( self ):
		return MpPrivate.motor_get_raw_acceleration_parameters( \
								self.handle )

	def set_raw_acceleration_parameters( self, parameter_list ):
		MpPrivate.motor_set_raw_acceleration_parameters( self.handle, \
								parameter_list )

	def get_acceleration_time( self ):
		return MpPrivate.motor_get_acceleration_time( self.handle )

	def set_acceleration_time( self ):
		MpPrivate.motor_set_acceleration_time( self.handle, \
							acceleration_time )

	def get_acceleration_distance( self ):
		return MpPrivate.motor_get_acceleration_distance( self.handle )

	def home_search_succeeded( self ):
		MpPrivate.motor_home_search_succeeded( self.handle )

	def get_gain( self, gain_type ):
		return MpPrivate.motor_get_gain( self.handle, gain_type )

	def set_gain( self, gain_type, gain ):
		MpPrivate.motor_set_gain( self.handle, gain_type, gain )

	def send_control_command( self, command_type, command ):
		MpPrivate.motor_send_control_command( self.handle,
						command_type, command )

#
#  Functions that operate on multiple motor objects.
#
def multiple_motors_move_absolute( motors_tuple, positions_tuple ):
	MpPrivate.multiple_motors_move_absolute( motors_tuple, positions_tuple )

def wait_for_multiple_motors_stop( motors_tuple ):
	MpPrivate.wait_for_multiple_motors_stop( motors_tuple )

#
#  Pan/Tilt/Zoom class
#

class PTZ(Record):
	def command( self, command ):
		MpPrivate.ptz_command( self.handle, command )

	def get_status( self ):
		return MpPrivate.ptz_get_status( self.handle )

	def pan_left( self ):
		MpPrivate.ptz_pan_left( self.handle )

	def pan_right( self ):
		MpPrivate.ptz_pan_right( self.handle )

	def pan_stop( self ):
		MpPrivate.ptz_pan_stop( self.handle )

	def get_pan( self ):
		return MpPrivate.ptz_get_pan( self.handle )

	def set_pan( self, pan_value ):
		MpPrivate.ptz_set_pan( self.handle, pan_value )

	def get_pan_speed( self ):
		return MpPrivate.ptz_get_pan_speed( self.handle )

	def set_pan_speed( self, pan_speed ):
		MpPrivate.ptz_set_pan_speed( self.handle, pan_speed )

	def tilt_up( self ):
		MpPrivate.ptz_tilt_up( self.handle )

	def tilt_down( self ):
		MpPrivate.ptz_tilt_down( self.handle )

	def tilt_stop( self ):
		MpPrivate.ptz_tilt_stop( self.handle )

	def get_tilt( self ):
		return MpPrivate.ptz_get_tilt( self.handle )

	def set_tilt( self, tilt_value ):
		MpPrivate.ptz_set_tilt( self.handle, tilt_value )

	def get_tilt_speed( self ):
		return MpPrivate.ptz_get_tilt_speed( self.handle )

	def set_tilt_speed( self, tilt_speed ):
		MpPrivate.ptz_set_tilt_speed( self.handle, tilt_speed )

	def zoom_in( self ):
		MpPrivate.ptz_zoom_in( self.handle )

	def zoom_out( self ):
		MpPrivate.ptz_zoom_out( self.handle )

	def zoom_stop( self ):
		MpPrivate.ptz_zoom_stop( self.handle )

	def get_zoom( self ):
		return MpPrivate.ptz_get_zoom( self.handle )

	def set_zoom( self, zoom_value ):
		MpPrivate.ptz_set_zoom( self.handle, zoom_value )

	def get_zoom_speed( self ):
		return MpPrivate.ptz_get_zoom_speed( self.handle )

	def set_zoom_speed( self, zoom_speed ):
		MpPrivate.ptz_set_zoom_speed( self.handle, zoom_speed )

	def zoom_off( self ):
		MpPrivate.ptz_zoom_off( self.handle )

	def zoom_on( self ):
		MpPrivate.ptz_zoom_on( self.handle )

	def focus_manual( self ):
		MpPrivate.ptz_focus_manual( self.handle )

	def focus_auto( self ):
		MpPrivate.ptz_focus_auto( self.handle )

	def focus_far( self ):
		MpPrivate.ptz_focus_far( self.handle )

	def focus_near( self ):
		MpPrivate.ptz_focus_near( self.handle )

	def focus_stop( self ):
		MpPrivate.ptz_focus_stop( self.handle )

	def get_focus( self ):
		return MpPrivate.ptz_get_focus( self.handle )

	def set_focus( self, focus_value ):
		MpPrivate.ptz_set_focus( self.handle, focus_value )

	def get_focus_speed( self ):
		return MpPrivate.ptz_get_focus_speed( self.handle )

	def set_focus_speed( self, focus_speed ):
		MpPrivate.ptz_set_focus_speed( self.handle, focus_speed )

#
#  Pulse Generator class
#

class PulseGenerator(Record):
	def is_busy( self ):
		return MpPrivate.pulse_generator_is_busy( self.handle )

	def start( self ):
		MpPrivate.pulse_generator_start( self.handle )

	def stop( self ):
		MpPrivate.pulse_generator_stop( self.handle )

	def get_mode( self ):
		return MpPrivate.pulse_generator_get_mode( self.handle )

	def set_mode( self, mode ):
		MpPrivate.pulse_generator_set_mode( self.handle, mode )

	def get_pulse_period( self ):
		return MpPrivate.pulse_generator_get_pulse_period( self.handle )

	def set_pulse_period( self, pulse_period ):
		MpPrivate.pulse_generator_set_pulse_period( self.handle,
								pulse_period )

	def get_pulse_width( self ):
		return MpPrivate.pulse_generator_get_pulse_width( self.handle )

	def set_pulse_width( self, pulse_width ):
		MpPrivate.pulse_generator_set_pulse_width( self.handle,
								pulse_width )

	def get_num_pulses( self ):
		return MpPrivate.pulse_generator_get_num_pulses( self.handle )

	def set_num_pulses( self, num_pulses ):
		MpPrivate.pulse_generator_set_num_pulses( self.handle,
								num_pulses )

	def get_pulse_delay( self ):
		return MpPrivate.pulse_generator_get_pulse_delay( self.handle )

	def set_pulse_delay( self, pulse_delay ):
		MpPrivate.pulse_generator_set_pulse_delay( self.handle,
								pulse_delay )


#
#  Relay class
#

class Relay(Record):
	def relay_command( self, command ):
		MpPrivate.relay_relay_command( self.handle, command )

	def get_relay_status( self ):
		return MpPrivate.relay_get_relay_status( self.handle )

#
#  Sample changer class
#

class SampleChanger(Record):
	def initialize( self ):
		MpPrivate.sample_changer_initialize( self.handle )

	def shutdown( self ):
		MpPrivate.sample_changer_shutdown( self.handle )

	def select_sample_holder( self, sample_holder_name ):
		MpPrivate.sample_changer_select_sample_holder( self.handle,
							   sample_holder_name )

	def unselect_sample_holder( self ):
		MpPrivate.sample_changer_unselect_sample_holder( self.handle )

	def grab_sample( self, sample_id ):
		MpPrivate.sample_changer_grab_sample( self.handle, sample_id )

	def ungrab_sample( self ):
		MpPrivate.sample_changer_ungrab_sample( self.handle )

	def mount_sample( self ):
		MpPrivate.sample_changer_mount_sample( self.handle )

	def unmount_sample( self ):
		MpPrivate.sample_changer_unmount_sample( self.handle )

	def get_sample_id( self ):
		return MpPrivate.sample_changer_get_sample_id( self.handle )

	def set_sample_id( self, sample_id ):
		MpPrivate.sample_changer_set_sample_id( self.handle, sample_id )

	def soft_abort( self ):
		MpPrivate.sample_changer_soft_abort( self.handle )

	def immediate_abort( self ):
		MpPrivate.sample_changer_immediate_abort( self.handle )

	def idle( self ):
		MpPrivate.sample_changer_idle( self.handle )

	def reset( self ):
		MpPrivate.sample_changer_reset( self.handle )

	def get_status( self ):
		return MpPrivate.sample_changer_get_status( self.handle )

	def is_busy( self ):
		return MpPrivate.sample_changer_is_busy( self.handle )

	def cooldown( self ):
		MpPrivate.sample_changer_cooldown( self.handle )

	def deice( self ):
		MpPrivate.sample_changer_deice( self.handle )

#
#  Single channel analyzer class
#

class SCA(Record):
	def get_lower_level( self ):
		return MpPrivate.sca_get_lower_level( self.handle )

	def set_lower_level( self, lower_level ):
		MpPrivate.sca_set_lower_level( self.handle, lower_level )

	def get_upper_level( self ):
		return MpPrivate.sca_get_upper_level( self.handle )

	def set_upper_level( self, upper_level ):
		MpPrivate.sca_set_upper_level( self.handle, upper_level )

	def get_gain( self ):
		MpPrivate.sca_get_gain( self.handle )

	def set_gain( self, gain ):
		MpPrivate.sca_set_gain( self.handle, gain )

	def get_time_constant( self ):
		MpPrivate.sca_get_time_constant( self.handle )

	def set_time_constant( self, time_constant ):
		MpPrivate.sca_set_time_constant( self.handle, time_constant )

	def get_mode( self ):
		MpPrivate.sca_get_mode( self.handle )

	def set_mode( self, mode ):
		MpPrivate.sca_set_mode( self.handle, mode )

#
#  Scaler class
#

class Scaler(Record):
	def clear( self ):
		MpPrivate.scaler_clear( self.handle )

	def overflow_set( self ):
		return MpPrivate.scaler_overflow_set( self.handle )

	def read( self ):
		return MpPrivate.scaler_read( self.handle )

	def read_raw( self ):
		return MpPrivate.scaler_read_raw( self.handle )

	def is_busy( self ):
		return MpPrivate.scaler_is_busy( self.handle )

	def start( self, preset_count ):
		MpPrivate.scaler_start( self.handle, preset_count )

	def stop( self ):
		return MpPrivate.scaler_stop( self.handle )

	def get_mode( self ):
		return MpPrivate.scaler_get_mode( self.handle )

	def set_mode( self, mode ):
		MpPrivate.scaler_set_mode( self.handle, mode )

	def set_modes_of_associated_counters( self ):
		MpPrivate.scaler_set_modes_of_associated_counters( self.handle )

	def get_dark_current( self ):
		return MpPrivate.scaler_get_dark_current( self.handle )

	def set_dark_current( self, dark_current ):
		MpPrivate.scaler_set_dark_current( self.handle, dark_current )

	def read_dark_current_subtracted_value( self, timer ):
		return MpPrivate.scaler_read_dark_current_subtracted_value( \
						self.handle, timer )

#
#  Table class
#

class Table(Record):

	def is_busy( self, axis_id ):
		return MpPrivate.table_is_busy( self.handle, axis_id )

	def move_absolute( self, axis_id, new_destination ):
		MpPrivate.table_move_absolute( self.handle, \
						axis_id, new_destination )

	def get_position( self, axis_id ):
		return MpPrivate.table_get_position( self.handle, axis_id )

	def set_position( self, axis_id, new_set_position ):
		MpPrivate.table_set_position( self.handle, \
						axis_id, new_set_position )

	def soft_abort( self, axis_id ):
		MpPrivate.table_soft_abort( self.handle, axis_id )

	def immediate_abort( self, axis_id ):
		MpPrivate.table_immediate_abort( self.handle, axis_id )

	def positive_limit_hit( self, axis_id ):
		return MpPrivate.table_positive_limit_hit( \
						self.handle, axis_id )

	def negative_limit_hit( self, axis_id ):
		return MpPrivate.table_negative_limit_hit( \
						self.handle, axis_id )

#
#  Timer class
#

class Timer(Record):
	def is_busy( self ):
		return MpPrivate.timer_is_busy( self.handle )

	def start( self, seconds ):
		MpPrivate.timer_start( self.handle, seconds )

	def stop( self ):
		return MpPrivate.timer_stop( self.handle )

	def clear( self ):
		MpPrivate.timer_clear( self.handle )

	def read( self ):
		return MpPrivate.timer_read( self.handle )

	def get_mode( self ):
		return MpPrivate.timer_get_mode( self.handle )

	def set_mode( self, mode ):
		MpPrivate.timer_set_mode( self.handle, mode )

	def set_modes_of_associated_counters( self ):
		MpPrivate.timer_set_modes_of_associated_counters( self.handle )

	def get_last_measurement_time( self ):
		return MpPrivate.timer_get_last_measurement_time( self.handle )

#
#  VideoInput class
#
#  FIXME: Implement the Python part of this class
#

class VideoInput(Record):
	pass

#
#  Scan class
#

class Scan(Record):
	def perform_scan( self ):
		MpPrivate.scan_perform_scan( self.handle )

#
#  Variable class
#

class Variable(Record):
	def read( self ):
		raw_value = MpPrivate.variable_receive_variable( self.handle )

		if ( type(raw_value) is types.TupleType ):
			if ( len(raw_value) == 1 ):
				new_value = raw_value[0]
			else:
				new_value = raw_value
		else:
			new_value = raw_value

		return new_value

	def write( self, value ):
		if ( type(value) is types.StringType ):
			new_value = value
		elif ( type(value) is types.TupleType ):
			new_value = value
		else:
			new_value = ( value, )

		MpPrivate.variable_send_variable( self.handle, new_value )

	def receive_variable( self ):
		return self.read()

	def send_variable( self, value ):
		self.write( value )

#
#  Server class
#

class Server(Record):
	def wait_for_message_id( self, message_id, timeout ):
		return None

	def wait_for_messages_from_server( self, timeout ):
		return MpPrivate.server_wait_for_messages_from_server( \
							self.handle, timeout)

#---------------------------------------------------------------------------
#
# FIXME: The following functions are inefficient and should be converted to C
#        or at least rewritten.  However, if the 'NumPy' module is imported,
#        these functions will never get used.  Since using 'NumPy' for large
#        arrays is the preferred mechanism, the inefficiency of the functions
#        below is not currently that critical.
#

import struct

#---------------------------------------------------------------------------

def mystr( x ):
	return str(x)

def myint( x ):
	return int(x)

def myfloat( x ):
	return float(x)

#---------------------------------------------------------------------------

def _get_struct_format_and_converter( mx_datatype ):

	if ( mx_datatype == 1 ):	# string
		fmt = "s"
		conv = mystr

	elif ( mx_datatype == 2 ):	# char
		fmt = "b"
		conv = myint

	elif ( mx_datatype == 3 ):	# unsigned char
		fmt = "B"
		conv = myint

	elif ( mx_datatype == 4 ):	# short
		fmt = "h"
		conv = myint

	elif ( mx_datatype == 5 ):	# unsigned short
		fmt = "H"
		conv = myint

	elif ( mx_datatype == 6 ):	# bool
		fmt = "i"
		conv = myint

	# Datatype 7 is reserved.

	elif ( mx_datatype == 8 ):	# long
		fmt = "l"
		conv = myint

	elif ( mx_datatype == 9 ):	# unsigned long
		fmt = "l"		#   Note: "L" produces a Python long.
		conv = myint

	elif ( mx_datatype == 10 ):	# float
		fmt = "f"
		conv = myfloat

	elif ( mx_datatype == 11 ):	# double
		fmt = "d"
		conv = myfloat

	elif ( mx_datatype == 12 ):	# hex
		fmt = "l"
		conv = myint

	# Datatype 13 is reserved.

	elif ( mx_datatype == 14 ):	# int64
		raise TypeError, "64-bit integers are not supported"

	elif ( mx_datatype == 15 ):	# uint64
		raise TypeError, "unsigned 64-bit integers are not supported"

	elif ( mx_datatype == 31 ):	# record name
		fmt = "s"
		conv = mystr

	elif ( mx_datatype == 32 ):	# recordtype name
		fmt = "s"
		conv = mystr

	elif ( mx_datatype == 33 ):	# interface name
		fmt = "s"
		conv = mystr

	else:
		error_message = "Unsupported MX datatype %d" % mx_datatype

		raise TypeError, error_message

	return ( fmt, conv )

#---------------------------------------------------------------------------

def _display_string( binary_string ):
	length = len(binary_string)

	for i in range(length):
		x = struct.unpack( "B", binary_string[i] )

		print "%#x " % ( x[0] ) ,

	print ""

#---------------------------------------------------------------------------

def _convert_binary_string_to_list( list, binary_string, datatype, \
			num_dimensions, dimension_tuple, element_size ):

	( fmt, conv ) = _get_struct_format_and_converter( datatype )

	if ( num_dimensions < 0 ):
		msg = "Illegal number of dimensions (%d)" % num_dimensions

		raise TypeError, msg

	elif ( num_dimensions == 0 ):
		value = struct.unpack( fmt, binary_string )

		if ( fmt != "s" ):
			value = value[0]

		list.append( value )

		return list

	elif (( num_dimensions == 1 ) and ( fmt == "s" )):

		#
		# 1-dimensional strings are special.  We must convert
		# the binary string to a null terminated string.
		#

		index = binary_string.find("\0")

		if ( index < 0 ):
			value = binary_string
		else:
			value = binary_string[:index]

		list.append( value )

		return list

	else:
		pass

	#
	# If we get here, then we either have a 2-dimensional (or above)
	# string array or a 1-dimensional (or above) array of some other
	# data type.
	#
	# This means that we must prepare to recursively call the function
	# _convert_binary_string_to_list() to handle the next level
	# of subarrays.

	subdimension_tuple = dimension_tuple[1:]

	num_subdimensions = len(subdimension_tuple)

	subarray_size = element_size

	for i in range(num_subdimensions):
		subarray_size = subarray_size * dimension_tuple[i+1]

	# Now we are ready to do the recursive calls.

	sublist = []

	for n in range( dimension_tuple[0] ):

		start = n * subarray_size
		end   = (n+1) * subarray_size

		substring = binary_string[start:end]

		sublist = _convert_binary_string_to_list( sublist, \
			substring, datatype, num_dimensions - 1,   \
			dimension_tuple[1:], element_size )

	list.append( sublist )

	return list

#---------------------------------------------------------------------------

def _flatten_sequence( myseq ):
	#
	# Expands a nested sequence into a 1-dimensional list.
	#
	# Warning: This function modifys the contents of deep elements of
	#          the original sequence.  If this is a problem, then you
	#          should do a deep copy before invoking this function.
	#

	mylist = list(myseq)	# Convert a top level tuple to a list.

	types = (list,tuple)

	try:
		for i in xrange(sys.maxint):

			#
			# If mylist[i] is a sequence type, we insert the
			# contents of that sequence in place.  By doing it
			# this way, the contents of the sequence will be at
			# the current index or later, which makes sure that
			# any embedded sequences will also be expanded.
			#
			# Along the way, tuples will be converted to lists.
			#

			while isinstance( mylist[i], types ):

				# Insert the sequence contents into the list.

				mylist[i:i+1] = list(mylist[i])

	except IndexError:
		#
		# When we get to the end of the expanded list, we will
		# get an IndexError exception for trying to use an index
		# value that is beyond the end of the list.
		#
		# We suppress that error and then return.
		#

		pass

	return mylist

#---------------------------------------------------------------------------

def _convert_sequence_to_binary_string( args, mx_datatype, num_dimensions, \
				dimension_tuple, element_size, field_name ):

	# Embed the arguments in a list, so that non-sequence args can be
	# handled along with sequence args.

	args = [ args ]

	# Convert the sequence to a 1-dimensional list.

	mylist = _flatten_sequence( args )

	# Figure out how to convert the sequence to a string.

	( fmt, conv ) = _get_struct_format_and_converter( mx_datatype )

	if ( fmt == "s" ):
		fmt_str = "%ds" % dimension_tuple[-1]

	# Compute the number of items that should be in 'mylist'.

	num_elements = 1

	if ( fmt == "s" ):
		# For strings, the last dimension is not included here.
		#
		for n in dimension_tuple[:-1]:
			num_elements = num_elements * n
	else:
		for n in dimension_tuple:
			num_elements = num_elements * n

	# Does the number of items match?

	if ( num_elements != len(mylist) ):
		error_message = "The number of items (%d) being assigned " \
			"to field '%s' does not match the actual number " \
			"of items (%d) in the field." \
				% ( len(mylist), field_name, num_elements )

		raise Illegal_Argument_Error, error_message

	#
	# Since strings are immutable, if we do something like this
	#
	#   binary_string = binary_string + "..."
	#
	# it will be very inefficient.  This is due to the fact that
	# each time we do the addition, a new binary string will be created
	# and the old one will be destroyed.
	#
	# Instead, we create a list of string fragments and do
	# a big join() operation at the end.
	# 

	string_list = []

	for x in mylist:
		cx = conv(x)

		if ( fmt == "s" ):
			string_fragment = struct.pack( fmt_str, cx )
		else:
			string_fragment = struct.pack( fmt, cx )

		string_list.append( string_fragment )

	binary_string = "".join( string_list )

	return binary_string

#---------------------------------------------------------------------------

#
#  Record field attribute definitions.
#

MXNA_VALUE_CHANGE_THRESHOLD = 1
MXNA_POLL                   = 2
MXNA_READ_ONLY              = 3

#---------------------------------------------------------------------------

#
#  Net class - Used to access individual network fields in a remote server.
#

class Net:
	def __init__( self, server_record, field_name ):
		self.server = server_record
		self.name  = field_name

		self.handle = \
			MpPrivate.net_init( server_record.handle, field_name )

	def __del__( self ):
		MpPrivate.net_del( self.handle )

	def get_field_parameters( self ):
		parameters = MpPrivate.net_get_field_parameters( self.handle )

		( self.datatype, self.num_dimensions, self.dimension_tuple, \
		    self.data_element_size_tuple, self.array_size ) = parameters

	def get( self ):
		if ( hasattr( self, "datatype" ) == 0 ):
			self.get_field_parameters()

		binary_string = \
			MpPrivate.net_get_remote_binary_string( self.handle )

		list = []

		list = _convert_binary_string_to_list(        \
					list,                 \
					binary_string,        \
					self.datatype,        \
					self.num_dimensions,  \
					self.dimension_tuple, \
					self.data_element_size_tuple[0] )

		if ( len(list) == 1 ):
			list = list[0]

		return list

	def put( self, args ):
		if ( hasattr( self, "datatype" ) == 0 ):
			self.get_field_parameters()

		binary_string = _convert_sequence_to_binary_string( \
					args,                       \
					self.datatype,              \
					self.num_dimensions,        \
					self.dimension_tuple,       \
					self.data_element_size_tuple[0], \
					self.name )

		MpPrivate.net_put_remote_binary_string( self.handle, \
							binary_string )

	def get_with_callback( self, \
			callback_function, callback_arguments, object=None ):

		# FIXME: The following kludge needs to be replaced with
		# an implementation that is _actually_ asynchronous.

		callback = Callback( None, \
			-1, callback_function, callback_arguments )

		callback.nf = self

		callback.object = object

		list = self.get()

		if ( callback_function != None ):
			callback_function( callback, callback_arguments )

		return list

	def get_attribute( self, attribute ):
		return MpPrivate.net_get_attribute( self.handle, attribute )

	def set_attribute( self, attribute, value ):
		MpPrivate.net_set_attribute( self.handle, attribute, value )

	def get_attribute_number( self, attribute_name ):
		return MpPrivate.net_get_attribute_number( self.handle, \
							attribute_name )

	def add_callback( self, callback_type, \
			callback_function, callback_arguments, object=None ):

		callback = Callback( None, \
			callback_type, callback_function, callback_arguments )

		callback.nf = self

		callback.object = object

		MpPrivate.net_add_callback(self.handle, callback)

		return callback

	def get_local_value( self ):
		if ( hasattr( self, "datatype" ) == 0 ):
			self.get_field_parameters()

		binary_string = \
			MpPrivate.net_get_local_binary_string( self.handle, \
							self.server.handle )

		list = []

		list = _convert_binary_string_to_list(        \
					list,                 \
					binary_string,        \
					self.datatype,        \
					self.num_dimensions,  \
					self.dimension_tuple, \
					self.data_element_size_tuple[0] )

		if ( len(list) == 1 ):
			list = list[0]

		return list

