/*
 * Name:    mp_net.c
 *
 * Purpose: Python module interface to generic MX net functions.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2006-2008 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MP_NET_DEBUG	FALSE

#include "Python.h"

#include "mx_util.h"
#include "mx_record.h"
#include "mx_net.h"
#include "mx_handle.h"
#include "mx_callback.h"
#include "mx_array.h"
#include "mp.h"

#include "mx_key.h"

/*-------------------------------------------------------------------------*/

static mx_status_type
MpPrivate_net_setup_field( long nf_handle,
			MX_NETWORK_FIELD **nf )
{
	static const char fname[] = "MpPrivate_net_setup_field()";

	void *nf_pointer;
	MX_RECORD_FIELD *local_field;
	long i, datatype, num_dimensions, num_data_element_sizes_to_allocate;
	long dimension_array[MXP_MAX_DIMENSIONS];
	size_t *sizeof_array;
	char *name_ptr;
	mx_bool_type new_handle_needed;
	mx_status_type mx_status;

	if ( nf == (MX_NETWORK_FIELD **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_NETWORK_FIELD pointer passed was NULL." );
	}

	mx_status = mx_get_pointer_from_handle( &nf_pointer,
				MpPrivate_global_network_field_handle_table,
				nf_handle );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	*nf = (MX_NETWORK_FIELD *) nf_pointer;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: nf_handle = %ld, nfname = '%s'",
		fname, nf_handle, (*nf)->nfname));
#endif

	/* Make sure that we have connected to the network field. */

	mx_status = mx_need_to_get_network_handle( *nf, &new_handle_needed );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( new_handle_needed ) {
		mx_status = mx_network_field_connect( *nf );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	/* We need to know the datatype and dimensions of the field.
	 * This information is stored in the MX_RECORD_FIELD structure
	 * which is pointed to by (*np)->local_field.
	 */

	if ( (*nf)->local_field != NULL ) {

		/* If this pointer is not NULL, then we assume that it
		 * has been set up correctly by a previous call.  If so,
		 * then we can return now.
		 */

		return MX_SUCCESSFUL_RESULT;
	}

	/* If we get here, we need to create and fill in the
	 * MX_RECORD_FIELD structure.
	 */

	local_field = calloc( 1, sizeof(MX_RECORD_FIELD) );

	if ( local_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an "
			"MX_RECORD_FIELD structure." );
	}

	(*nf)->local_field = local_field;

	(*nf)->must_free_local_field_on_delete = TRUE;

	local_field->flags = 0;		/* not MXFF_VARARGS */

	/* Find the field name for the remote field. */

	name_ptr = strrchr( (*nf)->nfname, '.' );

	if ( name_ptr == NULL ) {
		local_field->name = (*nf)->nfname;
	} else {
		local_field->name = name_ptr;
	}

	/* Get the field type parameters. */

	mx_status = mx_get_field_type( (*nf)->server_record, (*nf)->nfname,
					MXP_MAX_DIMENSIONS, &datatype,
					&num_dimensions, dimension_array );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Record, recordtype, and interface fields natively are pointers to
	 * the corresponding object.  We can't get at or use the pointer in
	 * the server, so we rewrite the network field type parameters to 
	 * refer to the ASCII name for the field type.  Since, the names
	 * are 1-dimensional strings, we must increase the number of 
	 * dimensions by 1 as well.
	 */

	if ( (datatype == MXFT_RECORD)
	  || (datatype == MXFT_RECORDTYPE)
	  || (datatype == MXFT_INTERFACE) )
	{
		num_dimensions++;

		switch( datatype ) {
		case MXFT_RECORD:
		  dimension_array[num_dimensions-1] = MXU_RECORD_NAME_LENGTH;
		  break;
		case MXFT_RECORDTYPE:
		  dimension_array[num_dimensions-1] = MXU_DRIVER_NAME_LENGTH;
		  break;
		case MXFT_INTERFACE:
		  dimension_array[num_dimensions-1] = MXU_INTERFACE_NAME_LENGTH;
		  break;
		}

		datatype = MXFT_STRING;
	}

	/* Copy the datatype and dimensions to the local field structure. */

	local_field->datatype = datatype;
	local_field->num_dimensions = num_dimensions;

	if ( num_dimensions == 0 ) {
		local_field->dimension = NULL;
	} else {
		local_field->dimension = malloc(num_dimensions * sizeof(long));

		if ( local_field->dimension == NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate a %ld element "
			"dimension array for server '%s', field '%s'.",
				num_dimensions,
				(*nf)->server_record->name,
				(*nf)->nfname );
		}
	}

	if ( num_dimensions == 0 ) {
		num_data_element_sizes_to_allocate = 1;
	} else {
		num_data_element_sizes_to_allocate = num_dimensions;
	}

	local_field->data_element_size
		= malloc( num_data_element_sizes_to_allocate * sizeof(size_t) );

	if ( local_field->data_element_size == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld element "
		"data element size array for server '%s', field '%s'.",
			num_data_element_sizes_to_allocate,
			(*nf)->server_record->name,
			(*nf)->nfname );
	}

	mx_status = mx_get_datatype_sizeof_array( datatype, &sizeof_array );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Save a copy of the dimension and sizeof arrays. */

	for ( i = 0; i < num_dimensions; i++ ) {
		local_field->dimension[i]         = dimension_array[i];
		local_field->data_element_size[i] = sizeof_array[i];
	}

	/* For a scalar, we always have 1 value in data_element_size_array. */

	if ( num_dimensions == 0 ) {
		local_field->data_element_size[0] = sizeof_array[0];
	}

	/* Allocate an array for the field data to be stored in. */

	if ( num_dimensions == 0 ) {
		local_field->data_pointer = malloc( sizeof_array[0] );
	} else {
		local_field->data_pointer = mx_allocate_array( num_dimensions,
						dimension_array, sizeof_array );
	}

	if ( local_field->data_pointer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory attempting to allocate a %ld dimensional "
		"array for server '%s', field '%s'.", num_dimensions,
			(*nf)->server_record->name, (*nf)->nfname );
	}

#if MP_NET_DEBUG
	{
		size_t array_size;

		/* Compute the total size of the data elements in the array. */

		mx_status = mx_get_array_size( local_field->num_dimensions,
					local_field->dimension,
					local_field->data_element_size,
					&array_size );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		MX_DEBUG(-2,("%s: server '%s', field '%s'",
			fname, (*nf)->server_record->name, (*nf)->nfname));

		MX_DEBUG(-2,("%s: datatype = %ld, num_dimensions = %ld",
			fname, local_field->datatype,
			local_field->num_dimensions));

		for ( i = 0; i < local_field->num_dimensions; i++ ) {
			MX_DEBUG(-2,("%s: dimension[%ld] = %ld",
				fname, i, local_field->dimension[i]));
		}

		MX_DEBUG(-2,("%s: array_size = %lu",
			fname, (unsigned long) array_size ));

		MX_DEBUG(-2,("%s: data_pointer = %p",
			fname, local_field->data_pointer));
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpPrivate_net_init_cmd( PyObject *self, PyObject *args )
{
	static const char fname[] = "MpPrivate_net_init_cmd()";

	PyObject *result;
	MX_RECORD *server_record;
	long server_handle, new_nf_handle;
	char *record_field_name;
	MX_NETWORK_FIELD *new_nf;
	int parse_status;
	mx_status_type mx_status;

	parse_status = PyArg_ParseTuple( args, "ls",
					&server_handle, &record_field_name );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_get_record_from_handle( server_handle,
							&server_record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for server '%s', record_field_name '%s'.",
		fname, server_record->name, record_field_name ));
#endif

	/* Allocate memory for a new MX_NETWORK_FIELD structure. */

	new_nf = malloc( sizeof(MX_NETWORK_FIELD) );

	if ( new_nf == (MX_NETWORK_FIELD *) NULL ) {
		MpPrivateErr_SetString( PyExc_MemoryError,
    "%s: Unable to allocate memory for an MX_NETWORK_FIELD structure.", fname );

		return NULL;
	}

	/* Initialize the contents of the MX_NETWORK_FIELD structure. */

	mx_status = mx_network_field_init( new_nf,
					server_record, record_field_name );

	if ( mx_status.code != MXE_SUCCESS ) {
		free( new_nf );
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* Create a handle for the MX_NETWORK_FIELD pointer. */

	mx_status = mx_create_handle( &new_nf_handle, 
				MpPrivate_global_network_field_handle_table,
				new_nf );

	if ( mx_status.code != MXE_SUCCESS ) {
		free( new_nf );
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* Send the network field handle back to Python. */

	result = Py_BuildValue( "l", new_nf_handle );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: new_nf_handle = %ld", fname, new_nf_handle));
#endif

	return result;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpPrivate_net_del_cmd( PyObject *self, PyObject *args )
{
#if MP_NET_DEBUG
	static const char fname[] = "MpPrivate_net_del_cmd()";
#endif

	MX_RECORD_FIELD *local_field;
	void *value_array;
	MX_NETWORK_FIELD *nf;
	long nf_handle;
	int python_status;
	mx_status_type mx_status;

	python_status = PyArg_ParseTuple( args, "l", &nf_handle );

	if ( python_status == 0 )
		return NULL;

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	local_field = nf->local_field;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for network field '%s', local_field = %p",
		fname, nf->nfname, local_field));
#endif

	if ( (local_field != NULL)
	  && (nf->must_free_local_field_on_delete) )
	{
		/* Free the array that the field data was stored in. */

		value_array = mx_get_field_value_pointer( local_field );

#if MP_NET_DEBUG
		MX_DEBUG(-2,("%s: value_array = %p", fname, value_array));
#endif

		if ( value_array != NULL ) {
			if ( local_field->num_dimensions == 0 ) {
				mx_free( value_array );
			} else {
				mx_free_array( value_array,
					local_field->num_dimensions,
					local_field->dimension,
					local_field->data_element_size );
			}
		}

		/* Free the other arrays in local_field. */

		mx_free( local_field->dimension );

		mx_free( local_field->data_element_size );

		/* Now free local_field itself. */

		mx_free( local_field );
	}

	/* Finish by freeing the MX_NETWORK_FIELD structure. */

	mx_free( nf );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s completed.", fname));
#endif

	Py_INCREF(Py_None);

	return Py_None;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpPrivate_net_get_field_parameters_cmd( PyObject *self, PyObject *args )
{
#if MP_NET_DEBUG
	static const char fname[] = "MpPrivate_net_get_field_parameters_cmd()";
#endif
	PyObject *result;
	PyObject *object;
	PyObject *dimension_tuple;
	PyObject *data_element_size_tuple;
	MX_NETWORK_FIELD *nf;
	MX_RECORD_FIELD *local_field;
	long i, tuple_length, nf_handle;
	size_t array_size;
	int python_status;
	mx_status_type mx_status;

	/* Get the MX_NETWORK_FIELD pointer from the supplied handle. */

	python_status = PyArg_ParseTuple( args, "l", &nf_handle );

	if ( python_status == 0 )
		return NULL;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for nf_handle = %ld", fname, nf_handle));
#endif

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	local_field = nf->local_field;

	/* Create a tuple to contain the dimension sizes. */

	dimension_tuple = PyTuple_New( local_field->num_dimensions );

	if ( dimension_tuple == NULL )
		return NULL;

	for ( i = 0; i < local_field->num_dimensions; i++ ) {
		object = PyInt_FromLong( local_field->dimension[i] );

		if ( object == NULL )
			return NULL;

		python_status = PyTuple_SetItem( dimension_tuple, i, object );

		if ( python_status != 0 )
			return NULL;
	}

	/* Create a tuple to contain the data element sizes.  The first
	 * element of the tuple will always be present, even for scalars,
	 * since we still need to know the size of the scalar.
	 */

	if ( local_field->num_dimensions == 0 ) {
		tuple_length = 1;
	} else {
		tuple_length = local_field->num_dimensions;
	}

	data_element_size_tuple = PyTuple_New( tuple_length );

	if ( data_element_size_tuple == NULL )
		return NULL;

	for ( i = 0; i < tuple_length; i++ ) {
		object = PyInt_FromLong( local_field->data_element_size[i] );

		if ( object == NULL )
			return NULL;

		python_status = PyTuple_SetItem( data_element_size_tuple,
							i, object );

		if ( python_status != 0 )
			return NULL;
	}

	mx_status = mx_get_array_size( local_field->num_dimensions,
					local_field->dimension,
					local_field->data_element_size,
					&array_size );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* Send the field parameters back to Python. */

	result = Py_BuildValue( "llOOl", local_field->datatype,
					local_field->num_dimensions,
					dimension_tuple,
					data_element_size_tuple,
					array_size );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s complete.", fname));
#endif

	return result;
}

/*-------------------------------------------------------------------------*/

static PyObject *
MpPrivate_net_get( PyObject *self,
		PyObject *args,
		MX_NETWORK_FIELD **nf,
		MX_RECORD_FIELD **local_field )
{
	static const char fname[] = "MpPrivate_net_get()";

	long nf_handle;
	int parse_status;
	void *value_array;
	mx_status_type mx_status;

	/* Get the MX_NETWORK_FIELD pointer from the supplied handle. */

	parse_status = PyArg_ParseTuple( args, "l", &nf_handle );

	if ( parse_status == 0 )
		return NULL;

	mx_status = MpPrivate_net_setup_field( nf_handle, nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for nf '%s'", fname, (*nf)->nfname));
#endif

	*local_field = (*nf)->local_field;

	if ( (*local_field) == NULL ) {
		mx_status = mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The local field pointer for network field '%s' is NULL.",
			(*nf)->nfname );

		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	value_array = mx_get_field_value_pointer( *local_field );

	if ( value_array == NULL ) {
		mx_status = mx_error( MXE_INITIALIZATION_ERROR, fname,
		"The data array for network field '%s' has not yet "
		"been allocated.", (*nf)->nfname );

		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* At this point, MpPrivate_net_setup_field() has already done
	 * everything that the top level mx_get_array() function would
	 * have done, so instead we call mx_internal_get_array() directly.
	 */

	mx_status = mx_internal_get_array( NULL, NULL, (*nf),
					(*local_field)->datatype,
					(*local_field)->num_dimensions,
					(*local_field)->dimension,
					value_array );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s complete", fname));
#endif

	Py_INCREF(Py_None);

	return Py_None;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/* The following routine returns the data to Python as a binary string
 * and is intended for use by NumPy.
 */

MX_EXPORT PyObject *
MpPrivate_net_get_remote_binary_string_cmd( PyObject *self, PyObject *args )
{
#if MP_NET_DEBUG
	static const char fname[] =
		"MpPrivate_net_get_remote_binary_string_cmd()";
#endif
	PyObject *result;
	PyObject *string_object;
	char *string_buffer;
	MX_NETWORK_FIELD *nf;
	MX_RECORD_FIELD *local_field;
	void *value_array;
	size_t value_array_size;
	mx_status_type mx_status;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked", fname));
#endif

	/* Suppress GCC 4 uninitialized variable warning. */

	local_field = NULL;

	/* Get the field value into the array local_field->data_array. */

	result = MpPrivate_net_get( self, args, &nf, &local_field );

	if ( result == NULL )
		return NULL;

	/*
	 * FIXME: When mx_allocate_array() has been changed to use a flat
	 *        data buffer, it will no longer be necessary to invoke
	 *        mx_copy_array_to_buffer() here.  Then we will be able
	 *        to just pass the pointer for the MX data buffer to
	 *        the function PyString_FromStringAndSize() directly.
	 *
	 *        This will be more efficient since it will effectively
	 *        eliminate one set of buffer copies.
	 */

	mx_status = mx_get_array_size( local_field->num_dimensions,
					local_field->dimension,
					local_field->data_element_size,
					&value_array_size );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	value_array = mx_get_field_value_pointer( local_field );

	/* Create an uninitialized Python string to contain the value(s)
	 * returned by MX.
	 */

	string_object = PyString_FromStringAndSize( NULL, value_array_size );

	if ( string_object == NULL )
		return NULL;

	/* Get a pointer to the new string buffer. */

	string_buffer = PyString_AsString( string_object );

	if ( string_buffer == NULL )
		return NULL;

	/* Now copy the MX array to the Python string buffer.
	 *
	 * See the FIXME above for why this should go away some day.
	 */

	if ( local_field->num_dimensions == 0 ) {

		/* 0-dimensional fields must be handled specially,
		 * since nf_private->dimension_array will be NULL
		 * and mx_copy_array_to_buffer() treats it as an
		 * error if any of its arguments is a NULL pointer.
		 *
		 * However, it is an easy case to handle.
		 */

		memcpy( string_buffer, value_array, value_array_size );
	} else {
		mx_status = mx_copy_array_to_buffer(
					value_array,
					TRUE,
					local_field->datatype,
					local_field->num_dimensions,
					local_field->dimension,
					local_field->data_element_size,
					string_buffer,
					value_array_size,
					NULL, FALSE );

		if ( mx_status.code != MXE_SUCCESS )
			return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s complete, string_object = %p", fname, string_object));
#endif

	return string_object;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/* The following routine returns the data as a potentially multi-dimensional
 * list and is intended for use by plain Python without NumPy.
 */

MX_EXPORT PyObject *
MpPrivate_net_get_object_cmd( PyObject *self, PyObject *args )
{
#if 1
	/* This function is not currently complete. */

	return NULL;
#else

#  if MP_NET_DEBUG
	static const char fname[] = "MpPrivate_net_get_object_cmd()";
#  endif

	PyObject *result, *new_sequence;
	MX_NETWORK_FIELD *nf;
	MX_RECORD_FIELD *local_field;
	mx_status_type mx_status;

	/* Suppress GCC 4 uninitialized variable warning. */

	local_field = NULL;

	/* Get the field value into the array nf_private->value_array. */

	result = MpPrivate_net_get( self, args, &nf, &local_field );

	if ( result == NULL )
		return NULL;

	/* Now create the list to send back to Python. */

	mx_status = MpPrivate_create_sequence_from_field( &local_field,
							MXP_ST_LIST,
							&new_sequence );
	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	return new_sequence;

#endif
}

/*-------------------------------------------------------------------------*/

/* The following routine receives data from Python as a binary string
 * and is intended for use by NumPy.
 */

MX_EXPORT PyObject *
MpPrivate_net_put_remote_binary_string_cmd( PyObject *self, PyObject *args )
{
	static const char fname[] =
		"MpPrivate_net_put_remote_binary_string_cmd()";

	MX_NETWORK_FIELD *nf;
	MX_RECORD_FIELD *local_field;
	long nf_handle;
	int parse_status, binary_string_length;
	void *value_array;
	size_t value_array_size;
	char *binary_string;
	mx_status_type mx_status;

	binary_string = NULL;	/* Suppress GCC 4 warning. */

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: self = %p, args = %p", fname, self, args));
#endif

	parse_status = PyArg_ParseTuple( args, "ls#",
			&nf_handle, &binary_string, &binary_string_length );

	if ( parse_status == 0 )
		return NULL;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: nf_handle = %ld, binary_string_length = %d",
			fname, nf_handle, binary_string_length));
#endif

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	local_field = nf->local_field;

	if ( local_field == NULL ) {
		mx_status = mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The local field pointer for network field '%s' is NULL.",
			nf->nfname );

		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	value_array = mx_get_field_value_pointer( local_field );

	if ( value_array == NULL ) {
		mx_status = mx_error( MXE_INITIALIZATION_ERROR, fname,
		"The data array for network field '%s' has not yet "
		"been allocated.", nf->nfname );

		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	mx_status = mx_get_array_size( local_field->num_dimensions,
					local_field->dimension,
					local_field->data_element_size,
					&value_array_size );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* For string types, make sure that Python's buffer will fit into
	 * MX's buffer.  For number types, make sure that Python's buffer
	 * and MX's buffer are the same size.
	 */

	switch( local_field->datatype ) {
	case MXFT_STRING:
		if ( binary_string_length > value_array_size ) {
			MpPrivateErr_SetString( MpPrivateExc_Would_Exceed_Limit,
				"The supplied Python string of length %ld "
				"is too long to fit into MX's buffer "
				"of length %ld for field '%s'.",
					binary_string_length,
					value_array_size, nf->nfname );
			return NULL;
		}
		break;

	case MXFT_RECORD:
	case MXFT_RECORDTYPE:
	case MXFT_INTERFACE:
		MpPrivateErr_SetString( MpPrivateExc_Illegal_Argument,
			"Cannot write to MX field '%s' since writing to "
			"'record', 'recordtype' and 'interface' fields "
			"is not supported.", nf->nfname );
		return NULL;
		break;

	case MXFT_CHAR:
	case MXFT_UCHAR:
	case MXFT_SHORT:
	case MXFT_USHORT:
	case MXFT_BOOL:
	case MXFT_LONG:
	case MXFT_ULONG:
	case MXFT_FLOAT:
	case MXFT_DOUBLE:
	case MXFT_HEX:
	case MXFT_INT64:
	case MXFT_UINT64:
		if ( binary_string_length != value_array_size ) {
			MpPrivateErr_SetString( MpPrivateExc_Type_Mismatch,
				"Python and MX do not agree on the length of "
				"the data buffer for field '%s'.  Python says "
				"that it should be %ld, while MX says that it "
				"should be %ld.  This should never happen.",
					nf->nfname, binary_string_length,
					value_array_size );

			return NULL;
		}
		break;

	default:
		MpPrivateErr_SetString( MpPrivateExc_Unsupported,
			"Cannot write to MX field '%s' since it has an "
			"unsupported datatype of %ld",
				nf->nfname, local_field->datatype );
		break;
	}

	/* Copy the data in the Python string into the MX value array.
	 *
	 * FIXME: Is there a way of doing this without copying the array?
	 */

	if ( local_field->num_dimensions == 0 ) {

		/* 0-dimensional fields must be handled specially,
		 * since nf_private->dimension_array will be NULL
		 * and mx_copy_array_to_buffer() treats it as an
		 * error if any of its arguments is a NULL pointer.
		 *
		 * However, it is an easy case to handle.
		 */

		memcpy( value_array, binary_string, value_array_size );
	} else {
#if 1
		/* For string values, the string length reported by Python
		 * does not include the space taken by the null terminator.
		 */

		if ( local_field->datatype == MXFT_STRING ) {
			binary_string_length++;
		}
#endif
		mx_status = mx_copy_buffer_to_array(
					binary_string,
					binary_string_length,
					value_array,
					TRUE,
					local_field->datatype,
					local_field->num_dimensions,
					local_field->dimension,
					local_field->data_element_size,
					NULL, FALSE );

		if ( mx_status.code != MXE_SUCCESS )
			return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	/* At this point, MpPrivate_net_setup_field() has already done
	 * everything that the top level mx_put_array() function would
	 * have done, so instead we call mx_internal_put_array() directly.
	 */

	mx_status = mx_internal_put_array( NULL, NULL, nf,
					local_field->datatype,
					local_field->num_dimensions,
					local_field->dimension,
					value_array );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: mx_internal_put_array() succeeded.", fname));
#endif

	Py_INCREF(Py_None);

	return Py_None;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

/* The following routine receives a sequence from Python and sends it on
 * to MX.  It is intended for use by plain Python without NumPy.
 */

MX_EXPORT PyObject *
MpPrivate_net_put_object_cmd( PyObject *self, PyObject *args )
{
	return NULL;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

MX_EXPORT PyObject *
MpPrivate_net_get_local_binary_string_cmd( PyObject *self, PyObject *args )
{
	static const char fname[] =
		"MpPrivate_net_get_local_binary_string_cmd()";

	PyObject *string_object;
	char *string_buffer;
	MX_NETWORK_FIELD *nf;
	MX_RECORD_FIELD *local_field;
	MX_RECORD *server_record;
	long nf_handle, server_handle;
	void *value_array;
	size_t value_array_size;
	int python_status;
	mx_status_type mx_status;

	python_status = PyArg_ParseTuple( args, "ll",
					&nf_handle, &server_handle );

	if ( python_status == 0 )
		return NULL;

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for nf '%s'", fname, nf->nfname));
#endif

	mx_status = MpPrivate_get_record_from_handle( server_handle,
							&server_record );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	local_field = nf->local_field;

	if ( local_field == NULL ) {
		mx_status = mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The local field pointer for network field '%s' is NULL.",
			nf->nfname );

		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	local_field->flags = 0;

	value_array = mx_get_field_value_pointer( local_field );

	if ( value_array == NULL ) {
		mx_status = mx_error( MXE_INITIALIZATION_ERROR, fname,
		"The data array for network field '%s' has not yet "
		"been allocated.", nf->nfname );

		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	mx_status = mx_get_array_size( local_field->num_dimensions,
					local_field->dimension,
					local_field->data_element_size,
					&value_array_size );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	/* Create an uninitialized Python string to contain the value(s)
	 * returned by MX.
	 */

	string_object = PyString_FromStringAndSize( NULL, value_array_size );

	if ( string_object == NULL )
		return NULL;

	/* Get a pointer to the new string buffer. */

	string_buffer = PyString_AsString( string_object );

	if ( string_buffer == NULL )
		return NULL;

	/* Now copy the MX array to the Python string buffer.
	 *
	 * See the FIXME in MpPrivate_net_get_binary_string_cmd()
	 * for why this should go away some day.
	 */

	if ( local_field->num_dimensions == 0 ) {

		/* 0-dimensional fields must be handled specially,
		 * since nf_private->dimension_array will be NULL
		 * and mx_copy_array_to_buffer() treats it as an
		 * error if any of its arguments is a NULL pointer.
		 *
		 * However, it is an easy case to handle.
		 */

		memcpy( string_buffer, value_array, value_array_size );
	} else {
		mx_status = mx_copy_array_to_buffer(
					value_array,
					TRUE,
					local_field->datatype,
					local_field->num_dimensions,
					local_field->dimension,
					local_field->data_element_size,
					string_buffer,
					value_array_size,
					NULL, FALSE );

		if ( mx_status.code != MXE_SUCCESS )
			return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s complete, string_object = %p", fname, string_object));
#endif

	return string_object;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

MX_EXPORT PyObject *
MpPrivate_net_get_attribute_cmd( PyObject *self, PyObject *args )
{
#if MP_NET_DEBUG
	static const char fname[] = "MpPrivate_net_get_attribute_cmd()";
#endif

	PyObject *result;
	PyObject *attribute_object;
	MX_NETWORK_FIELD *nf;
	long nf_handle;
	int python_status;
	unsigned long attribute_number;
	char *attribute_name;
	double attribute_value;
	mx_status_type mx_status;

	python_status = PyArg_ParseTuple( args, "lO",
					&nf_handle, &attribute_object );

	if ( python_status == 0 )
		return NULL;

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for network field '%s'.",
		fname, nf->nfname));
#endif

	if ( PyInt_Check(attribute_object) ) {
		/* If we were passed an integer, then we assume that
		 * it is the attribute number.
		 */

		attribute_number = PyInt_AsLong( attribute_object );

#if MP_NET_DEBUG
		MX_DEBUG(-2,("%s: attribute number = %lu",
				fname, attribute_number));
#endif
		mx_status = mx_network_field_get_attribute( nf,
							attribute_number,
							&attribute_value );
	} else
	if ( PyLong_Check(attribute_object) ) {
		/* If we were passed a long, then we assume that
		 * it is the attribute number.
		 */

		attribute_number = PyLong_AsLong( attribute_object );

#if MP_NET_DEBUG
		MX_DEBUG(-2,("%s: attribute number = %lu",
				fname, attribute_number));
#endif
		mx_status = mx_network_field_get_attribute( nf,
							attribute_number,
							&attribute_value );
	} else
	if ( PyString_Check(attribute_object) ) {
		/* If we were passed a string, then we assume that
		 * it is the attribute name.
		 */

		attribute_name = PyString_AsString( attribute_object );

#if MP_NET_DEBUG
		MX_DEBUG(-2,("%s: attribute name = '%s'",
				fname, attribute_name));
#endif
		mx_status = mx_network_field_get_attribute_by_name( nf,
							attribute_name,
							&attribute_value );
	} else {
		MpPrivateErr_SetString( MpPrivateExc_Type_Mismatch,
			"The supplied Python attribute '%s' is not "
			"a Python integer or a Python string for field '%s'.",
				PyObject_Repr(attribute_object), nf->nfname );
		return NULL;
	}

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: attribute_value = %g", fname, attribute_value));
#endif

	result = Py_BuildValue( "d", attribute_value );

	return result;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

MX_EXPORT PyObject *
MpPrivate_net_set_attribute_cmd( PyObject *self, PyObject *args )
{
#if MP_NET_DEBUG
	static const char fname[] = "MpPrivate_net_set_attribute_cmd()";
#endif

	PyObject *attribute_object;
	MX_NETWORK_FIELD *nf;
	long nf_handle;
	int python_status;
	unsigned long attribute_number;
	char *attribute_name;
	double attribute_value;
	mx_status_type mx_status;

	python_status = PyArg_ParseTuple( args, "lOd",
			&nf_handle, &attribute_object, &attribute_value );

	if ( python_status == 0 )
		return NULL;

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for network field '%s'.",
		fname, nf->nfname));
#endif

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: attribute_value = %g", fname, attribute_value));
#endif

	if ( PyInt_Check(attribute_object) ) {
		/* If we were passed an integer, then we assume that
		 * it is the attribute number.
		 */

		attribute_number = PyInt_AsLong( attribute_object );

#if MP_NET_DEBUG
		MX_DEBUG(-2,("%s: attribute number = %lu",
				fname, attribute_number));
#endif
		mx_status = mx_network_field_set_attribute( nf,
							attribute_number,
							attribute_value );
	} else
	if ( PyLong_Check(attribute_object) ) {
		/* If we were passed a long, then we assume that
		 * it is the attribute number.
		 */

		attribute_number = PyLong_AsLong( attribute_object );

#if MP_NET_DEBUG
		MX_DEBUG(-2,("%s: attribute number = %lu",
				fname, attribute_number));
#endif
		mx_status = mx_network_field_set_attribute( nf,
							attribute_number,
							attribute_value );
	} else
	if ( PyString_Check(attribute_object) ) {
		/* If we were passed a string, then we assume that
		 * it is the attribute name.
		 */

		attribute_name = PyString_AsString( attribute_object );

#if MP_NET_DEBUG
		MX_DEBUG(-2,("%s: attribute name = '%s'",
				fname, attribute_name));
#endif
		mx_status = mx_network_field_get_attribute_by_name( nf,
							attribute_name,
							&attribute_value );
	} else {
		MpPrivateErr_SetString( MpPrivateExc_Type_Mismatch,
			"The supplied Python attribute '%s' is not "
			"a Python integer or a Python string for field '%s'.",
				PyObject_Repr(attribute_object), nf->nfname );
		return NULL;
	}

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s completed.", fname));
#endif

	Py_INCREF(Py_None);

	return Py_None;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

MX_EXPORT PyObject *
MpPrivate_net_get_attribute_number_cmd( PyObject *self, PyObject *args )
{
#if MP_NET_DEBUG
	static const char fname[] = "MpPrivate_net_get_attribute_number_cmd()";
#endif

	PyObject *result;
	MX_NETWORK_FIELD *nf;
	long nf_handle;
	int python_status;
	unsigned long attribute_number;
	char *attribute_name;
	mx_status_type mx_status;

	python_status = PyArg_ParseTuple( args, "ls",
					&nf_handle, &attribute_name );

	if ( python_status == 0 )
		return NULL;

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for network field '%s', attribute name '%s'.",
		fname, nf->nfname, attribute_name));
#endif

	mx_status = mx_network_field_get_attribute_number( nf,
							attribute_name,
							&attribute_number );
#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: attribute_number = %lu", fname, attribute_number));
#endif

	result = Py_BuildValue( "l", attribute_number );

	return result;
}

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

static mx_status_type
MpPrivate_net_callback_function( MX_CALLBACK *callback,
				void *arguments )
{
	static const char fname[] = "MpPrivate_net_callback_function()";

	PyObject *callback_object;
	PyObject *callback_function_object;
	PyObject *callback_arguments_object;
	PyObject *result;
	PyObject *arglist;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: callback_id = %lu",
		fname, (unsigned long) callback->callback_id));
#endif

	callback_object = arguments;

	callback_function_object = PyObject_GetAttrString( callback_object,
							"callback_function" );

	if ( callback_function_object == NULL ) {

#if ( PY_MAJOR_VERSION >= 2 )
		PyErr_WriteUnraisable( callback_object );
#endif
		PyErr_Clear();

		return mx_error( MXE_CORRUPT_DATA_STRUCTURE | MXE_QUIET, fname,
		"The callback object passed did not have 'callback_function' "
		"as an attribute." );
	}

	callback_arguments_object = PyObject_GetAttrString( callback_object,
							"callback_arguments" );

	if ( callback_arguments_object == NULL ) {

#if ( PY_MAJOR_VERSION >= 2 )
		PyErr_WriteUnraisable( callback_object );
#endif
		PyErr_Clear();

		return mx_error( MXE_CORRUPT_DATA_STRUCTURE | MXE_QUIET, fname,
		"The callback object passed did not have 'callback_arguments' "
		"as an attribute." );
	}

	arglist = Py_BuildValue("(OO)",
			callback_object, callback_arguments_object);

	if ( arglist == NULL ) {

#if ( PY_MAJOR_VERSION >= 2 )
		PyErr_WriteUnraisable( callback_function_object );
#endif
		PyErr_Clear();

		return mx_error( MXE_FUNCTION_FAILED | MXE_QUIET, fname,
		"Wrapping the callback arguments in a tuple failed." );
	}

	result = PyEval_CallObject( callback_function_object, arglist );

	Py_DECREF( arglist );

	if ( result == NULL ) {

#if ( PY_MAJOR_VERSION >= 2 )
		PyErr_WriteUnraisable( callback_function_object );
#endif
		PyErr_Clear();

		return mx_error( MXE_FUNCTION_FAILED | MXE_QUIET, fname,
		"Calling the Python callback failed." );
	}

	Py_DECREF( result );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT PyObject *
MpPrivate_net_add_callback_cmd( PyObject *self, PyObject *args )
{
	static const char fname[] = "MpPrivate_net_add_callback_cmd()";

	PyObject *callback_object;
	PyObject *callback_type_attribute;
	PyObject *callback_id_attribute;
	PyObject *callback_address_attribute;
	MX_NETWORK_FIELD *nf;
	MX_CALLBACK *callback;
	long nf_handle, callback_type;
	int python_status;
	mx_status_type mx_status;

	python_status = PyArg_ParseTuple( args, "lO",
					&nf_handle, &callback_object );
	if ( python_status == 0 )
		return NULL;

	if ( callback_object == Py_None ) {
		mx_status = mx_error( MXE_CORRUPT_DATA_STRUCTURE | MXE_QUIET,
		fname, "The callback object passed was None." );

		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );
	}

	mx_status = MpPrivate_net_setup_field( nf_handle, &nf );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for network field '%s'.", fname, nf->nfname));
#endif

	/* Get the callback type from the callback object. */

	callback_type_attribute = PyObject_GetAttrString( callback_object,
							"callback_type" );

	if ( callback_type_attribute == NULL )
		return NULL;

	callback_type = PyLong_AsLong( callback_type_attribute );

	Py_DECREF( callback_type_attribute );

	/* Setup the callback in the C library. */

	mx_status = mx_remote_field_add_callback( nf, callback_type,
						MpPrivate_net_callback_function,
						callback_object,
						&callback );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

	if ( callback == (MX_CALLBACK *) NULL ) {
		MpPrivateErr_SetString( PyExc_RuntimeError,
	"The MX_CALLBACK pointer returned by mx_remote_field_add_callback() "
	"for network field '%s' was NULL.", nf->nfname );

		return NULL;
	}

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: field '%s', callback_id = %lu",
		fname, nf->nfname, (unsigned long) callback->callback_id));
#endif

	/* Set the callback_id attribute in the callback object. */

	callback_id_attribute = PyLong_FromLong( callback->callback_id );

	if ( callback_id_attribute == NULL )
		return NULL;

	python_status = PyObject_SetAttrString( callback_object,
						"callback_id",
						callback_id_attribute );
	Py_DECREF( callback_id_attribute );

	if ( python_status == (-1) )
		return NULL;

	/* Add the address of the C MX_CALLBACK structure as
	 * a private attribute.
	 */

	/* FIXME: Casting a pointer to an unsigned long is BAD! */

	callback_address_attribute = PyLong_FromLong((unsigned long) callback);

	if ( callback_address_attribute == NULL )
		return NULL;

	python_status = PyObject_SetAttrString( callback_object,
						"__callback_address",
						callback_address_attribute );

	Py_DECREF( callback_address_attribute );

	if ( python_status == (-1) )
		return NULL;

	/* We are done, so return now. */

	Py_INCREF(Py_None);

	return Py_None;
}

MX_EXPORT PyObject *
MpPrivate_net_del_callback_cmd( PyObject *self, PyObject *args )
{
#if MP_NET_DEBUG
	static const char fname[] = "MpPrivate_net_del_callback_cmd()";
#endif
	PyObject *callback_object;
	PyObject *callback_address_attribute;
	MX_CALLBACK *callback;
	int python_status;
	mx_status_type mx_status;

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s invoked for self = %p, args = %p",
		fname, self, args));
#endif

	python_status = PyArg_ParseTuple( args, "O", &callback_object );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: callback_object = %p", fname, callback_object));
#endif

	if ( python_status == 0 )
		return NULL;

	/* Get the address of the C MX_CALLBACK structure from
	 * the object.
	 */

	callback_address_attribute = PyObject_GetAttrString( callback_object,
							"__callback_address" );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: callback_address_attribute = %p",
		fname, callback_address_attribute));
#endif

	if ( callback_address_attribute == NULL )
		return NULL;

	/* FIXME: Casting an unsigned long to a pointer is BAD! */

	callback = (void *) PyLong_AsUnsignedLong( callback_address_attribute );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s: C callback = %p", fname, callback));
#endif

	/* Call the C library to delete the callback. */

	mx_status = mx_delete_callback( callback );

	if ( mx_status.code != MXE_SUCCESS )
		return MpPrivateErr_SetFromMxStatus( NULL, mx_status );

#if MP_NET_DEBUG
	MX_DEBUG(-2,("%s complete", fname));
#endif

	Py_INCREF(Py_None);

	return Py_None;
}

